# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for expression generation
#

import os
import platform
import sys

import pyomo.common.unittest as unittest

from pyomo.environ import (
    ConcreteModel,
    RangeSet,
    Set,
    Param,
    Var,
    Expression,
    ExternalFunction,
    VarList,
    sum_product,
    inequality,
    quicksum,
    sin,
    tanh,
    value,
)
from pyomo.core.expr.numvalue import native_types, nonpyomo_leaf_types, NumericConstant
from pyomo.core.expr.numeric_expr import (
    SumExpression,
    ProductExpression,
    MonomialTermExpression,
    LinearExpression,
    NPV_SumExpression,
    NPV_ProductExpression,
    NegationExpression,
    NPV_NegationExpression,
    PowExpression,
    NPV_PowExpression,
    MaxExpression,
    NPV_MaxExpression,
    MinExpression,
    NPV_MinExpression,
    DivisionExpression,
    NPV_DivisionExpression,
    UnaryFunctionExpression,
    NPV_UnaryFunctionExpression,
    AbsExpression,
    NPV_AbsExpression,
)
from pyomo.core.expr.visitor import (
    FixedExpressionError,
    NonConstantExpressionError,
    SimpleExpressionVisitor,
    StreamBasedExpressionVisitor,
    ExpressionReplacementVisitor,
    evaluate_expression,
    expression_to_string,
    replace_expressions,
    sizeof_expression,
    identify_variables,
    identify_components,
    identify_mutable_parameters,
    RECURSION_LIMIT,
    get_stack_depth,
)
from pyomo.core.base.param import ParamData, ScalarParam
from pyomo.core.expr.template_expr import IndexTemplate
from pyomo.common.collections import ComponentSet
from pyomo.common.errors import TemplateExpressionError
from pyomo.common.log import LoggingIntercept
from io import StringIO
from pyomo.core.expr.compare import assertExpressionsEqual


class TestExpressionUtilities(unittest.TestCase):
    def test_identify_vars_numeric(self):
        #
        # There are no variables in a constant expression
        #
        self.assertEqual(list(identify_variables(5)), [])

    def test_identify_vars_params(self):
        m = ConcreteModel()
        m.I = RangeSet(3)
        m.a = Param(initialize=1)
        m.b = Param(m.I, initialize=1, mutable=True)
        #
        # There are no variables in expressions with only parameters
        #
        self.assertEqual(list(identify_variables(m.a)), [])
        self.assertEqual(list(identify_variables(m.b[1])), [])
        self.assertEqual(list(identify_variables(m.a + m.b[1])), [])
        self.assertEqual(list(identify_variables(m.a ** m.b[1])), [])
        self.assertEqual(list(identify_variables(m.a ** m.b[1] + m.b[2])), [])
        self.assertEqual(
            list(identify_variables(m.a ** m.b[1] + m.b[2] * m.b[3] * m.b[2])), []
        )

    def test_identify_duplicate_vars(self):
        #
        # Identify variables when there are duplicates
        #
        m = ConcreteModel()
        m.a = Var(initialize=1)

        # self.assertEqual( list(identify_variables(2*m.a+2*m.a, allow_duplicates=True)),
        #                  [ m.a, m.a ] )
        self.assertEqual(list(identify_variables(2 * m.a + 2 * m.a)), [m.a])

    def test_identify_vars_expr(self):
        #
        # Identify variables in named expressions
        #
        m = ConcreteModel()
        m.a = Var(initialize=1)
        m.b = Var(initialize=2)
        m.e = Expression(expr=3 * m.a)
        m.E = Expression([0, 1], initialize={0: 3 * m.a, 1: 4 * m.b})

        self.assertEqual(list(identify_variables(m.b + m.e)), [m.b, m.a])
        self.assertEqual(list(identify_variables(m.E[0])), [m.a])
        self.assertEqual(list(identify_variables(m.E[1])), [m.b])

    def test_identify_vars_expr_cache(self):
        #
        # Identify variables in named expressions
        #
        m = ConcreteModel()
        m.a = Var(initialize=1)
        m.b = Var(initialize=2)
        m.c = Var(initialize=3)
        m.d = Var(initialize=4)
        m.e = Expression(expr=3 * m.a)

        cache = {}
        self.assertEqual(
            list(identify_variables(m.b + m.e, named_expression_cache=cache)),
            [m.b, m.a],
        )
        self.assertEqual(cache, {id(m.e): ({id(m.a): m.a}, {id(m.e): (m.e, m.e.expr)})})

        # Check that the cache is used (to check, we will cause the cache to lie)
        s, e = cache[id(m.e)]
        s.clear()
        s.update({id(m.b): m.b, id(m.c): m.c})
        self.assertEqual(
            list(identify_variables(m.b + m.e, named_expression_cache=cache)),
            [m.b, m.c],
        )

        # Check that changing the expression invalidates the cache
        m.e = 4 * m.d
        self.assertEqual(
            list(identify_variables(m.b + m.e, named_expression_cache=cache)),
            [m.b, m.d],
        )

        # Check that changing a nested expression invalidates the cache
        m.f = Expression(expr=5 * m.a * m.e * m.b)
        self.assertEqual(
            list(identify_variables(m.c + m.f, named_expression_cache=cache)),
            [m.c, m.a, m.d, m.b],
        )
        self.assertEqual(
            cache,
            {
                id(m.e): ({id(m.d): m.d}, {id(m.e): (m.e, m.e.expr)}),
                id(m.f): (
                    {id(m.a): m.a, id(m.d): m.d, id(m.b): m.b},
                    {id(m.f): (m.f, m.f.expr), id(m.e): (m.e, m.e.expr)},
                ),
            },
        )
        m.e = 5
        self.assertEqual(
            list(identify_variables(m.c + m.f, named_expression_cache=cache)),
            [m.c, m.a, m.b],
        )
        self.assertEqual(
            cache,
            {
                id(m.e): ({}, {id(m.e): (m.e, m.e.expr)}),
                id(m.f): (
                    {id(m.a): m.a, id(m.b): m.b},
                    {id(m.f): (m.f, m.f.expr), id(m.e): (m.e, m.e.expr)},
                ),
            },
        )

    def test_identify_vars_vars(self):
        m = ConcreteModel()
        m.I = RangeSet(3)
        m.a = Var(initialize=1)
        m.b = Var(m.I, initialize=1)
        m.p = Param(initialize=1, mutable=True)
        m.x = ExternalFunction(library='foo.so', function='bar')
        #
        # Identify variables in various algebraic expressions
        #
        self.assertEqual(list(identify_variables(m.a)), [m.a])
        self.assertEqual(list(identify_variables(m.b[1])), [m.b[1]])
        self.assertEqual(list(identify_variables(m.a + m.b[1])), [m.a, m.b[1]])
        self.assertEqual(list(identify_variables(m.a ** m.b[1])), [m.a, m.b[1]])
        self.assertEqual(
            ComponentSet(identify_variables(m.a ** m.b[1] + m.b[2])),
            ComponentSet([m.b[2], m.a, m.b[1]]),
        )
        self.assertEqual(
            list(identify_variables(m.a ** m.b[1] + m.b[2] * m.b[3] * m.b[2])),
            [m.a, m.b[1], m.b[2], m.b[3]],
        )
        self.assertEqual(
            list(identify_variables(m.a ** m.b[1] + m.b[2] / m.b[3] * m.b[2])),
            [m.a, m.b[1], m.b[2], m.b[3]],
        )
        #
        # Identify variables in the arguments to functions
        #
        self.assertEqual(
            ComponentSet(identify_variables(m.x(m.a, 'string_param', 1, []) * m.b[1])),
            ComponentSet([m.b[1], m.a]),
        )
        self.assertEqual(
            list(identify_variables(m.x(m.p, 'string_param', 1, []) * m.b[1])), [m.b[1]]
        )
        self.assertEqual(
            ComponentSet(identify_variables(tanh(m.a) * m.b[1])),
            ComponentSet([m.b[1], m.a]),
        )
        self.assertEqual(
            ComponentSet(identify_variables(abs(m.a) * m.b[1])),
            ComponentSet([m.b[1], m.a]),
        )
        #
        # Check logic for allowing duplicates
        #
        self.assertEqual(list(identify_variables(m.a**m.a + m.a)), [m.a])
        # self.assertEqual( list(identify_variables(m.a**m.a + m.a, allow_duplicates=True)),
        #                  [ m.a, m.a, m.a,  ] )

    def test_identify_vars_linear_expression(self):
        m = ConcreteModel()
        m.x = Var()
        expr = quicksum([m.x, m.x], linear=True)
        self.assertEqual(list(identify_variables(expr, include_fixed=False)), [m.x])


class TestIdentifyParams(unittest.TestCase):
    def test_identify_params_numeric(self):
        #
        # There are no parameters in a constant expression
        #
        self.assertEqual(list(identify_mutable_parameters(5)), [])

    def test_identify_mutable_parameters(self):
        m = ConcreteModel()
        m.I = RangeSet(3)
        m.a = Var(initialize=1)
        m.b = Var(m.I, initialize=1)
        #
        # There are no parameters in expressions with only vars
        #
        self.assertEqual(list(identify_mutable_parameters(m.a)), [])
        self.assertEqual(list(identify_mutable_parameters(m.b[1])), [])
        self.assertEqual(list(identify_mutable_parameters(m.a + m.b[1])), [])
        self.assertEqual(list(identify_mutable_parameters(m.a ** m.b[1])), [])
        self.assertEqual(list(identify_mutable_parameters(m.a ** m.b[1] + m.b[2])), [])
        self.assertEqual(
            list(identify_mutable_parameters(m.a ** m.b[1] + m.b[2] * m.b[3] * m.b[2])),
            [],
        )

    def test_identify_mutable_parameters_constants(self):
        #
        # ScalarParams and NumericConstants are not recognized
        #
        m = ConcreteModel()
        m.x = Var(initialize=1)
        m.x.fix()
        m.p = Param(initialize=2, mutable=False)
        m.p_m = Param(initialize=3, mutable=True)
        e1 = m.x + m.p + NumericConstant(5)
        self.assertEqual(list(identify_mutable_parameters(e1)), [])

        e2 = 5 * m.x + NumericConstant(3) * m.p_m + m.p == 0
        mut_params = list(identify_mutable_parameters(e2))
        self.assertEqual(len(mut_params), 1)
        self.assertIs(mut_params[0], m.p_m)

    def test_identify_duplicate_params(self):
        #
        # Identify mutable params when there are duplicates
        #
        m = ConcreteModel()
        m.a = Param(initialize=1, mutable=True)

        self.assertEqual(list(identify_mutable_parameters(2 * m.a + 2 * m.a)), [m.a])

    def test_identify_mutable_parameters_expr(self):
        #
        # Identify mutable params in named expressions
        #
        m = ConcreteModel()
        m.a = Param(initialize=1, mutable=True)
        m.b = Param(initialize=2, mutable=True)
        m.e = Expression(expr=3 * m.a)
        m.E = Expression([0, 1], initialize={0: 3 * m.a, 1: 4 * m.b})

        self.assertEqual(list(identify_mutable_parameters(m.b + m.e)), [m.b, m.a])
        self.assertEqual(list(identify_mutable_parameters(m.E[0])), [m.a])
        self.assertEqual(list(identify_mutable_parameters(m.E[1])), [m.b])

    def test_identify_mutable_parameters_logical_expr(self):
        #
        # Identify mutable params in logical expressions
        #
        m = ConcreteModel()
        m.a = Param(initialize=0, mutable=True)
        expr = m.a + 1 == 0
        param_set = ComponentSet(identify_mutable_parameters(expr))
        self.assertEqual(len(param_set), 1)
        self.assertIn(m.a, param_set)

    def test_identify_mutable_parameters_params(self):
        m = ConcreteModel()
        m.I = RangeSet(3)
        m.a = Param(initialize=1, mutable=True)
        m.b = Param(m.I, initialize=1, mutable=True)
        m.p = Var(initialize=1)
        m.x = ExternalFunction(library='foo.so', function='bar')
        #
        # Identify variables in various algebraic expressions
        #
        self.assertEqual(list(identify_mutable_parameters(m.a)), [m.a])
        self.assertEqual(list(identify_mutable_parameters(m.b[1])), [m.b[1]])
        self.assertEqual(list(identify_mutable_parameters(m.a + m.b[1])), [m.a, m.b[1]])
        self.assertEqual(
            list(identify_mutable_parameters(m.a ** m.b[1])), [m.a, m.b[1]]
        )
        self.assertEqual(
            list(identify_mutable_parameters(m.a ** m.b[1] + m.b[2])),
            [m.a, m.b[1], m.b[2]],
        )
        self.assertEqual(
            list(identify_mutable_parameters(m.a ** m.b[1] + m.b[2] * m.b[3] * m.b[2])),
            [m.a, m.b[1], m.b[2], m.b[3]],
        )
        self.assertEqual(
            list(identify_mutable_parameters(m.a ** m.b[1] + m.b[2] / m.b[3] * m.b[2])),
            [m.a, m.b[1], m.b[2], m.b[3]],
        )
        #
        # Identify variables in the arguments to functions
        #
        self.assertEqual(
            list(identify_mutable_parameters(m.x(m.a, 'string_param', 1, []) * m.b[1])),
            [m.a, m.b[1]],
        )
        self.assertEqual(
            list(identify_mutable_parameters(m.x(m.p, 'string_param', 1, []) * m.b[1])),
            [m.b[1]],
        )
        self.assertEqual(
            list(identify_mutable_parameters(tanh(m.a) * m.b[1])), [m.a, m.b[1]]
        )
        self.assertEqual(
            list(identify_mutable_parameters(abs(m.a) * m.b[1])), [m.a, m.b[1]]
        )
        #
        # Check logic for allowing duplicates
        #
        self.assertEqual(list(identify_mutable_parameters(m.a**m.a + m.a)), [m.a])


#
# Replace all variables with new variables from a varlist
#
class ReplacementWalkerTest1(ExpressionReplacementVisitor):
    def __init__(self, model):
        ExpressionReplacementVisitor.__init__(self)
        self.model = model

    def visiting_potential_leaf(self, node):
        if node.__class__ in nonpyomo_leaf_types or not node.is_potentially_variable():
            return True, node
        if node.is_variable_type():
            if id(node) in self.substitute:
                return True, self.substitute[id(node)]
            self.substitute[id(node)] = self.model.w.add()
            return True, self.substitute[id(node)]
        return False, None


class WalkerTests(unittest.TestCase):
    def test_replacement_walker1(self):
        M = ConcreteModel()
        M.x = Var()
        M.y = Var()
        M.w = VarList()

        e = sin(M.x) + M.x * M.y + 3
        walker = ReplacementWalkerTest1(M)
        f = walker.dfs_postorder_stack(e)
        assertExpressionsEqual(self, sin(M.x) + M.x * M.y + 3, e)
        assertExpressionsEqual(self, sin(M.w[1]) + M.w[1] * M.w[2] + 3, f)

    def test_replacement_walker2(self):
        M = ConcreteModel()
        M.x = Var()
        M.w = VarList()

        e = M.x
        walker = ReplacementWalkerTest1(M)
        f = walker.dfs_postorder_stack(e)
        assertExpressionsEqual(self, M.x, e)
        assertExpressionsEqual(self, M.w[1], f)

    def test_replacement_walker3(self):
        M = ConcreteModel()
        M.x = Var()
        M.y = Var()
        M.w = VarList()

        e = sin(M.x) + M.x * M.y + 3 <= 0
        walker = ReplacementWalkerTest1(M)
        f = walker.dfs_postorder_stack(e)
        assertExpressionsEqual(self, sin(M.x) + M.x * M.y + 3 <= 0, e)
        assertExpressionsEqual(self, sin(M.w[1]) + M.w[1] * M.w[2] + 3 <= 0, f)

    def test_replacement_walker4(self):
        M = ConcreteModel()
        M.x = Var()
        M.y = Var()
        M.w = VarList()

        e = inequality(0, sin(M.x) + M.x * M.y + 3, 1)
        walker = ReplacementWalkerTest1(M)
        f = walker.dfs_postorder_stack(e)
        assertExpressionsEqual(self, inequality(0, sin(M.x) + M.x * M.y + 3, 1), e)
        assertExpressionsEqual(
            self, inequality(0, sin(M.w[1]) + M.w[1] * M.w[2] + 3, 1), f
        )

    def test_replacement_walker0(self):
        M = ConcreteModel()
        M.x = Var(range(3))
        M.w = VarList()
        M.z = Param(range(3), mutable=True)

        e = sum_product(M.z, M.x)
        self.assertIs(type(e), LinearExpression)
        walker = ReplacementWalkerTest1(M)
        f = walker.dfs_postorder_stack(e)
        assertExpressionsEqual(
            self,
            LinearExpression(
                linear_coefs=[i for i in M.z.values()],
                linear_vars=[i for i in M.x.values()],
            ),
            e,
        )
        assertExpressionsEqual(
            self,
            LinearExpression(
                linear_coefs=[i for i in M.z.values()],
                linear_vars=[i for i in M.w.values()],
            ),
            f,
        )

        del M.w
        M.w = VarList()
        e = 2 * sum_product(M.z, M.x)
        walker = ReplacementWalkerTest1(M)
        f = walker.dfs_postorder_stack(e)
        assertExpressionsEqual(
            self,
            2
            * LinearExpression(
                linear_coefs=[i for i in M.z.values()],
                linear_vars=[i for i in M.x.values()],
            ),
            e,
        )
        assertExpressionsEqual(
            self,
            2
            * LinearExpression(
                linear_coefs=[i for i in M.z.values()],
                linear_vars=[i for i in M.w.values()],
            ),
            f,
        )

    def test_replacement_linear_expression_with_constant(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        e = LinearExpression(linear_coefs=[2], linear_vars=[m.x])
        e += m.y
        sub_map = dict()
        sub_map[id(m.x)] = 5
        e2 = replace_expressions(e, sub_map)
        assertExpressionsEqual(self, e2, LinearExpression([10, m.y]))

        e = LinearExpression(linear_coefs=[2, 3], linear_vars=[m.x, m.y])
        sub_map = dict()
        sub_map[id(m.x)] = 5
        e2 = replace_expressions(e, sub_map)
        assertExpressionsEqual(
            self, e2, LinearExpression([10, MonomialTermExpression((3, m.y))])
        )

    def test_replacement_linear_expression_with_nonlinear(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        e = LinearExpression(linear_coefs=[2, 3], linear_vars=[m.x, m.y])
        sub_map = dict()
        sub_map[id(m.x)] = m.x**2
        e2 = replace_expressions(e, sub_map)
        assertExpressionsEqual(self, e2, SumExpression([2 * m.x**2, 3 * m.y]))

    def test_replace_expressions_with_monomial_term(self):
        M = ConcreteModel()
        M.x = Var()
        e = 2.0 * M.x
        substitution_map = {id(M.x): 3.0 * M.x}
        new_e = replace_expressions(e, substitution_map=substitution_map)
        self.assertEqual('6.0*x', str(new_e))
        # See comment about this test in ExpressionReplacementVisitor
        # old code would print '2.0*3.0*x'

    def test_identify_components(self):
        M = ConcreteModel()
        M.x = Var()
        M.y = Var()
        M.w = Var()

        e = sin(M.x) + M.x * M.w + 3
        v = list(str(v) for v in identify_components(e, set([M.x.__class__])))
        self.assertEqual(v, ['x', 'w'])
        v = list(str(v) for v in identify_components(e, [M.x.__class__]))
        self.assertEqual(v, ['x', 'w'])

    def test_identify_variables(self):
        M = ConcreteModel()
        M.x = Var()
        M.y = Var()
        M.w = Var()
        M.w = 2
        M.w.fixed = True

        e = sin(M.x) + M.x * M.w + 3
        v = list(str(v) for v in identify_variables(e))
        self.assertEqual(v, ['x', 'w'])
        v = list(str(v) for v in identify_variables(e, include_fixed=False))
        self.assertEqual(v, ['x'])

    def test_expression_to_string(self):
        M = ConcreteModel()
        M.x = Var()
        M.w = Var()

        e = sin(M.x) + M.x * M.w + 3
        self.assertEqual("sin(x) + x*w + 3", expression_to_string(e))
        M.w = 2
        M.w.fixed = True
        self.assertEqual(
            "sin(x) + x*2 + 3", expression_to_string(e, compute_values=True)
        )

    def test_expression_component_to_string(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        m.e = Expression(expr=m.x * m.y)
        m.f = Expression(expr=m.e)

        e = m.x + m.f * m.y
        self.assertEqual("x + ((x*y))*y", str(e))
        self.assertEqual("x + ((x*y))*y", expression_to_string(e))


#
# Replace all variables with a product expression
#
class ReplacementWalkerTest2(ExpressionReplacementVisitor):
    def __init__(self, model):
        ExpressionReplacementVisitor.__init__(self)
        self.model = model

    def visiting_potential_leaf(self, node):
        if node.__class__ in nonpyomo_leaf_types or not node.is_potentially_variable():
            return True, node

        if node.is_variable_type():
            if id(node) in self.substitute:
                return True, self.substitute[id(node)]
            self.substitute[id(node)] = 2 * self.model.w.add()
            return True, self.substitute[id(node)]
        return False, None


class WalkerTests2(unittest.TestCase):
    def test_replacement_walker1(self):
        M = ConcreteModel()
        M.x = Var()
        M.y = Var()
        M.w = VarList()

        e = sin(M.x) + M.x * M.y + 3
        walker = ReplacementWalkerTest2(M)
        f = walker.dfs_postorder_stack(e)
        assertExpressionsEqual(self, sin(M.x) + M.x * M.y + 3, e)
        assertExpressionsEqual(self, sin(2 * M.w[1]) + 2 * M.w[1] * (2 * M.w[2]) + 3, f)

    def test_replacement_walker2(self):
        M = ConcreteModel()
        M.x = Var()
        M.w = VarList()

        e = M.x
        walker = ReplacementWalkerTest2(M)
        f = walker.dfs_postorder_stack(e)
        assertExpressionsEqual(self, M.x, e)
        assertExpressionsEqual(self, 2 * M.w[1], f)

    def test_replacement_walker3(self):
        M = ConcreteModel()
        M.x = Var()
        M.y = Var()
        M.w = VarList()

        e = sin(M.x) + M.x * M.y + 3 <= 0
        walker = ReplacementWalkerTest2(M)
        f = walker.dfs_postorder_stack(e)
        assertExpressionsEqual(self, sin(M.x) + M.x * M.y + 3 <= 0, e)
        assertExpressionsEqual(
            self, sin(2 * M.w[1]) + 2 * M.w[1] * (2 * M.w[2]) + 3 <= 0, f
        )

    def test_replacement_walker4(self):
        M = ConcreteModel()
        M.x = Var()
        M.y = Var()
        M.w = VarList()

        e = inequality(0, sin(M.x) + M.x * M.y + 3, 1)
        walker = ReplacementWalkerTest2(M)
        f = walker.dfs_postorder_stack(e)
        assertExpressionsEqual(self, inequality(0, sin(M.x) + M.x * M.y + 3, 1), e)
        assertExpressionsEqual(
            self, inequality(0, sin(2 * M.w[1]) + 2 * M.w[1] * (2 * M.w[2]) + 3, 1), f
        )

    def test_replacement_walker5(self):
        M = ConcreteModel()
        M.x = Var()
        M.w = VarList()
        M.z = Param(mutable=True)

        e = M.z * M.x
        walker = ReplacementWalkerTest2(M)
        f = walker.dfs_postorder_stack(e)
        assertExpressionsEqual(self, e, MonomialTermExpression((M.z, M.x)))
        assertExpressionsEqual(
            self, f, MonomialTermExpression((NPV_ProductExpression((M.z, 2)), M.w[1]))
        )

    def test_replacement_walker0(self):
        M = ConcreteModel()
        M.x = Var(range(3))
        M.w = VarList()
        M.z = Param(range(3), mutable=True)

        e = sum_product(M.z, M.x)
        self.assertIs(type(e), LinearExpression)
        walker = ReplacementWalkerTest2(M)
        f = walker.dfs_postorder_stack(e)
        assertExpressionsEqual(
            self,
            e,
            LinearExpression(
                linear_coefs=[i for i in M.z.values()],
                linear_vars=[i for i in M.x.values()],
            ),
        )
        assertExpressionsEqual(
            self,
            f,
            LinearExpression(
                [
                    MonomialTermExpression(
                        (NPV_ProductExpression((M.z[0], 2)), M.w[1])
                    ),
                    MonomialTermExpression(
                        (NPV_ProductExpression((M.z[1], 2)), M.w[2])
                    ),
                    MonomialTermExpression(
                        (NPV_ProductExpression((M.z[2], 2)), M.w[3])
                    ),
                ]
            ),
        )

        e = 2 * sum_product(M.z, M.x)
        walker = ReplacementWalkerTest2(M)
        f = walker.dfs_postorder_stack(e)
        assertExpressionsEqual(
            self,
            2
            * LinearExpression(
                linear_coefs=[i for i in M.z.values()],
                linear_vars=[i for i in M.x.values()],
            ),
            e,
        )
        assertExpressionsEqual(
            self,
            f,
            ProductExpression(
                (
                    2,
                    LinearExpression(
                        [
                            MonomialTermExpression(
                                (NPV_ProductExpression((M.z[0], 2)), M.w[4])
                            ),
                            MonomialTermExpression(
                                (NPV_ProductExpression((M.z[1], 2)), M.w[5])
                            ),
                            MonomialTermExpression(
                                (NPV_ProductExpression((M.z[2], 2)), M.w[6])
                            ),
                        ]
                    ),
                )
            ),
        )


#
# Replace all mutable parameters with variables
#
class ReplacementWalkerTest3(ExpressionReplacementVisitor):
    def __init__(self, model):
        super().__init__(remove_named_expressions=False)
        self.model = model

    def visiting_potential_leaf(self, node):
        if node.__class__ in (ParamData, ScalarParam):
            if id(node) in self.substitute:
                return True, self.substitute[id(node)]
            self.substitute[id(node)] = 2 * self.model.w.add()
            return True, self.substitute[id(node)]

        if (
            node.__class__ in nonpyomo_leaf_types
            or node.is_constant()
            or node.is_variable_type()
        ):
            return True, node

        return False, None


class WalkerTests3(unittest.TestCase):
    def test_replacement_walker1(self):
        M = ConcreteModel()
        M.x = Param(mutable=True)
        M.y = Var()
        M.w = VarList()

        e = sin(M.x) + M.x * M.y + 3
        walker = ReplacementWalkerTest3(M)
        f = walker.dfs_postorder_stack(e)
        assertExpressionsEqual(self, sin(M.x) + M.x * M.y + 3, e)
        assertExpressionsEqual(self, sin(2 * M.w[1]) + 2 * M.w[1] * M.y + 3, f)

    def test_replacement_walker2(self):
        M = ConcreteModel()
        M.x = Param(mutable=True)
        M.w = VarList()

        e = M.x
        walker = ReplacementWalkerTest3(M)
        f = walker.dfs_postorder_stack(e)
        assertExpressionsEqual(self, M.x, e)
        assertExpressionsEqual(self, 2 * M.w[1], f)

    def test_replacement_walker3(self):
        M = ConcreteModel()
        M.x = Param(mutable=True)
        M.y = Var()
        M.w = VarList()

        e = sin(M.x) + M.x * M.y + 3 <= 0
        walker = ReplacementWalkerTest3(M)
        f = walker.dfs_postorder_stack(e)
        assertExpressionsEqual(self, sin(M.x) + M.x * M.y + 3 <= 0, e)
        assertExpressionsEqual(self, sin(2 * M.w[1]) + 2 * M.w[1] * M.y + 3 <= 0, f)

    def test_replacement_walker4(self):
        M = ConcreteModel()
        M.x = Param(mutable=True)
        M.y = Var()
        M.w = VarList()

        e = inequality(0, sin(M.x) + M.x * M.y + 3, 1)
        walker = ReplacementWalkerTest3(M)
        f = walker.dfs_postorder_stack(e)
        assertExpressionsEqual(self, inequality(0, sin(M.x) + M.x * M.y + 3, 1), e)
        assertExpressionsEqual(
            self, inequality(0, sin(2 * M.w[1]) + 2 * M.w[1] * M.y + 3, 1), f
        )

    def test_replacement_walker5(self):
        M = ConcreteModel()
        M.x = Var()
        M.w = VarList()
        M.z = Param(mutable=True)

        e = M.z * M.x
        walker = ReplacementWalkerTest3(M)
        f = walker.dfs_postorder_stack(e)
        self.assertIs(e.__class__, MonomialTermExpression)
        self.assertIs(f.__class__, ProductExpression)
        self.assertTrue(f.arg(0).is_potentially_variable())
        assertExpressionsEqual(self, M.z * M.x, e)
        assertExpressionsEqual(self, 2 * M.w[1] * M.x, f)

    def test_replacement_walker6(self):
        M = ConcreteModel()
        M.x = Var()
        M.w = VarList()
        M.z = Param(mutable=True)

        e = (M.z * 2) * 3
        walker = ReplacementWalkerTest3(M)
        f = walker.dfs_postorder_stack(e)
        self.assertTrue(not e.is_potentially_variable())
        self.assertTrue(f.is_potentially_variable())
        assertExpressionsEqual(self, M.z * 2 * 3, e)
        assertExpressionsEqual(
            self, ProductExpression([ProductExpression([2 * M.w[1], 2]), 3]), f
        )

    def test_replacement_walker7(self):
        M = ConcreteModel()
        M.x = Var()
        M.w = VarList()
        M.z = Param(mutable=True)
        M.e = Expression(expr=M.z * 2)

        e = M.x * M.e
        self.assertTrue(e.arg(1).is_potentially_variable())
        self.assertTrue(not e.arg(1).arg(0).is_potentially_variable())
        assertExpressionsEqual(
            self,
            ProductExpression([M.x, (NPV_ProductExpression([M.z, 2]))]),
            e,
            include_named_exprs=False,
        )
        walker = ReplacementWalkerTest3(M)
        f = walker.dfs_postorder_stack(e)
        self.assertTrue(e.__class__ is ProductExpression)
        self.assertTrue(f.__class__ is ProductExpression)
        self.assertEqual(id(e), id(f))
        self.assertTrue(f.arg(1).is_potentially_variable())
        self.assertTrue(f.arg(1).arg(0).is_potentially_variable())
        assertExpressionsEqual(
            self, M.x * ProductExpression([2 * M.w[1], 2]), f, include_named_exprs=False
        )

    def test_replacement_walker0(self):
        M = ConcreteModel()
        M.x = Var(range(3))
        M.w = VarList()
        M.z = Param(range(3), mutable=True)

        e = sum_product(M.z, M.x)
        self.assertIs(type(e), LinearExpression)
        walker = ReplacementWalkerTest3(M)
        f = walker.dfs_postorder_stack(e)
        assertExpressionsEqual(
            self,
            LinearExpression(
                linear_coefs=[i for i in M.z.values()],
                linear_vars=[i for i in M.x.values()],
            ),
            e,
        )
        assertExpressionsEqual(
            self, 2 * M.w[1] * M.x[0] + 2 * M.w[2] * M.x[1] + 2 * M.w[3] * M.x[2], f
        )

        e = 2 * sum_product(M.z, M.x)
        walker = ReplacementWalkerTest3(M)
        f = walker.dfs_postorder_stack(e)
        assertExpressionsEqual(
            self,
            2
            * LinearExpression(
                linear_coefs=[i for i in M.z.values()],
                linear_vars=[i for i in M.x.values()],
            ),
            e,
        )
        assertExpressionsEqual(
            self,
            2 * (2 * M.w[4] * M.x[0] + 2 * M.w[5] * M.x[1] + 2 * M.w[6] * M.x[2]),
            f,
        )


#
# Replace all mutable parameters with variables
#
class ReplacementWalker_ReplaceInternal(ExpressionReplacementVisitor):
    def exitNode(self, node, data):
        if type(node) == ProductExpression:
            return sum(data[1])
        else:
            return super().exitNode(node, data)


class WalkerTests_ReplaceInternal(unittest.TestCase):
    def test_no_replacement(self):
        m = ConcreteModel()
        m.x = Param(mutable=True)
        m.y = Var([1, 2, 3])

        e = sum(m.y[i] for i in m.y) == 0
        f = ReplacementWalker_ReplaceInternal().dfs_postorder_stack(e)
        assertExpressionsEqual(self, m.y[1] + m.y[2] + m.y[3] == 0, e)
        assertExpressionsEqual(self, m.y[1] + m.y[2] + m.y[3] == 0, f)
        self.assertIs(e, f)

    def test_replace(self):
        m = ConcreteModel()
        m.x = Param(mutable=True)
        m.y = Var([1, 2, 3])

        e = m.y[1] * m.y[2] + m.y[2] * m.y[3] == 0
        f = ReplacementWalker_ReplaceInternal().dfs_postorder_stack(e)
        assertExpressionsEqual(self, m.y[1] * m.y[2] + m.y[2] * m.y[3] == 0, e)
        assertExpressionsEqual(
            self,
            SumExpression(
                [LinearExpression([m.y[1], m.y[2]]), LinearExpression([m.y[2], m.y[3]])]
            )
            == 0,
            f,
        )

    def test_replace_nested(self):
        m = ConcreteModel()
        m.x = Param(mutable=True)
        m.y = Var([1, 2, 3])

        e = m.y[1] * m.y[2] * m.y[2] * m.y[3] == 0
        f = ReplacementWalker_ReplaceInternal().dfs_postorder_stack(e)
        assertExpressionsEqual(self, m.y[1] * m.y[2] * m.y[2] * m.y[3] == 0, e)
        assertExpressionsEqual(self, m.y[1] + m.y[2] + m.y[2] + m.y[3] == 0, f)
        self.assertIs(type(f.arg(0)), LinearExpression)
        self.assertEqual(f.arg(0).nargs(), 4)


class TestReplacementWithNPV(unittest.TestCase):
    def test_npv_sum(self):
        m = ConcreteModel()
        m.p1 = Param(mutable=True)
        m.p2 = Param(mutable=True)
        m.x = Var()

        e1 = m.p1 + 2
        e2 = replace_expressions(e1, {id(m.p1): m.p2})
        e3 = replace_expressions(e1, {id(m.p1): m.x})

        assertExpressionsEqual(self, e2, m.p2 + 2)
        assertExpressionsEqual(self, e3, LinearExpression([m.x, 2]))

    def test_npv_negation(self):
        m = ConcreteModel()
        m.p1 = Param(mutable=True)
        m.p2 = Param(mutable=True)
        m.x = Var()

        e1 = -m.p1
        e2 = replace_expressions(e1, {id(m.p1): m.p2})
        e3 = replace_expressions(e1, {id(m.p1): m.x})

        assertExpressionsEqual(self, e2, -m.p2)
        assertExpressionsEqual(self, e3, NegationExpression([m.x]))

    def test_npv_pow(self):
        m = ConcreteModel()
        m.p1 = Param(mutable=True)
        m.p2 = Param(mutable=True)
        m.x = Var()

        e1 = m.p1**3
        e2 = replace_expressions(e1, {id(m.p1): m.p2})
        e3 = replace_expressions(e1, {id(m.p1): m.x})

        assertExpressionsEqual(self, e2, m.p2**3)
        assertExpressionsEqual(self, e3, m.x**3)

    def test_npv_product(self):
        m = ConcreteModel()
        m.p1 = Param(mutable=True)
        m.p2 = Param(mutable=True)
        m.x = Var()

        e1 = m.p1 * 3
        e2 = replace_expressions(e1, {id(m.p1): m.p2})
        e3 = replace_expressions(e1, {id(m.p1): m.x})

        assertExpressionsEqual(self, e2, m.p2 * 3)
        assertExpressionsEqual(self, e3, ProductExpression([m.x, 3]))

    def test_npv_div(self):
        m = ConcreteModel()
        m.p1 = Param(mutable=True)
        m.p2 = Param(mutable=True)
        m.x = Var()

        e1 = m.p1 / 3
        e2 = replace_expressions(e1, {id(m.p1): m.p2})
        e3 = replace_expressions(e1, {id(m.p1): m.x})

        assertExpressionsEqual(self, e2, m.p2 / 3)
        assertExpressionsEqual(self, e3, DivisionExpression((m.x, 3)))

    def test_npv_unary(self):
        m = ConcreteModel()
        m.p1 = Param(mutable=True)
        m.p2 = Param(mutable=True)
        m.x = Var(initialize=0)

        e1 = sin(m.p1)
        e2 = replace_expressions(e1, {id(m.p1): m.p2})
        e3 = replace_expressions(e1, {id(m.p1): m.x})

        assertExpressionsEqual(self, e2, sin(m.p2))
        assertExpressionsEqual(self, e3, sin(m.x))

    def test_npv_abs(self):
        m = ConcreteModel()
        m.p1 = Param(mutable=True)
        m.p2 = Param(mutable=True)
        m.x = Var()

        e1 = abs(m.p1)
        e2 = replace_expressions(e1, {id(m.p1): m.p2})
        e3 = replace_expressions(e1, {id(m.p1): m.x})

        assertExpressionsEqual(self, e2, abs(m.p2))
        assertExpressionsEqual(self, e3, abs(m.x))


class BaseStreamBasedVisitorTests:
    def setUp(self):
        self.m = m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        m.z = Var()
        # Note: we do not use the operator overloading to generate the
        # expression so that the structure is constant even when we make
        # adjustments to the expression generators
        self.e = SumExpression(
            [
                PowExpression((m.x, 2)),
                m.y,
                ProductExpression((m.z, SumExpression([m.x, m.y]))),
            ]
        )

    def test_bad_args(self):
        with self.assertRaisesRegex(
            RuntimeError, "Unrecognized keyword arguments: {'foo': None}"
        ):
            StreamBasedExpressionVisitor(foo=None)

    def test_default(self):
        walker = StreamBasedExpressionVisitor()
        ans = self.walk(walker, self.e)
        ref = [[[], []], [], [[], [[], []]]]
        self.assertEqual(ans, ref)

    def test_beforeChild(self):
        def before(node, child, child_idx):
            if type(child) in nonpyomo_leaf_types or not child.is_expression_type():
                return False, [child]

        walker = StreamBasedExpressionVisitor(beforeChild=before)
        ans = self.walk(walker, self.e)
        m = self.m
        ref = [[[m.x], [2]], [m.y], [[m.z], [[m.x], [m.y]]]]
        self.assertEqual(str(ans), str(ref))

        ans = self.walk(walker, m.x)
        ref = []
        self.assertEqual(str(ans), str(ref))

        ans = self.walk(walker, 2)
        ref = []
        self.assertEqual(str(ans), str(ref))

    def test_initializeWalker_beforeChild(self):
        def before(node, child, child_idx):
            if type(child) in nonpyomo_leaf_types or not child.is_expression_type():
                return False, child

        def initialize(expr):
            ans = before(None, expr, 0)
            if ans is None:
                return True, expr
            else:
                return ans

        walker = StreamBasedExpressionVisitor(
            beforeChild=before, initializeWalker=initialize
        )
        ans = self.walk(walker, self.e)
        m = self.m
        ref = [[m.x, 2], m.y, [m.z, [m.x, m.y]]]
        self.assertEqual(str(ans), str(ref))

        ans = self.walk(walker, m.x)
        ref = m.x
        self.assertEqual(str(ans), str(ref))

        ans = self.walk(walker, 2)
        ref = 2
        self.assertEqual(str(ans), str(ref))

    def test_beforeChild_exitNode(self):
        def before(node, child, child_idx):
            if type(child) in nonpyomo_leaf_types or not child.is_expression_type():
                return False, [child]

        def exit(node, data):
            if hasattr(node, 'getname'):
                data.insert(0, node.getname())
            else:
                data.insert(0, str(node))
            return data

        walker = StreamBasedExpressionVisitor(beforeChild=before, exitNode=exit)
        ans = self.walk(walker, self.e)
        m = self.m
        ref = [
            'sum',
            ['pow', [m.x], [2]],
            [m.y],
            ['prod', [m.z], ['sum', [m.x], [m.y]]],
        ]
        self.assertEqual(str(ans), str(ref))

        ans = self.walk(walker, m.x)
        ref = ['x']
        self.assertEqual(str(ans), str(ref))

        ans = self.walk(walker, 2)
        ref = ['2']
        self.assertEqual(str(ans), str(ref))

    def test_beforeChild_enterNode_exitNode(self):
        i = [0]

        def before(node, child, child_idx):
            if type(child) in nonpyomo_leaf_types or not child.is_expression_type():
                return False, [child]

        def enter(node):
            i[0] += 1
            return None, [i[0]]

        def exit(node, data):
            if hasattr(node, 'getname'):
                data.insert(0, node.getname())
            else:
                data.insert(0, str(node))
            return data

        walker = StreamBasedExpressionVisitor(
            beforeChild=before, enterNode=enter, exitNode=exit
        )
        ans = self.walk(walker, self.e)
        m = self.m
        ref = [
            'sum',
            1,
            ['pow', 2, [m.x], [2]],
            [m.y],
            ['prod', 3, [m.z], ['sum', 4, [m.x], [m.y]]],
        ]
        self.assertEqual(str(ans), str(ref))

        ans = self.walk(walker, m.x)
        ref = ['x', 5]
        self.assertEqual(str(ans), str(ref))

        ans = self.walk(walker, 2)
        ref = ['2', 6]
        self.assertEqual(str(ans), str(ref))

    def test_old_beforeChild(self):
        def before(node, child):
            if type(child) in nonpyomo_leaf_types or not child.is_expression_type():
                return False, [child]

        os = StringIO()
        with LoggingIntercept(os, 'pyomo'):
            walker = StreamBasedExpressionVisitor(beforeChild=before)
        self.assertIn(
            "Note that the API for the StreamBasedExpressionVisitor "
            "has changed to include the child index for the beforeChild() "
            "method",
            os.getvalue().replace('\n', ' '),
        )

        ans = self.walk(walker, self.e)
        m = self.m
        ref = [[[m.x], [2]], [m.y], [[m.z], [[m.x], [m.y]]]]
        self.assertEqual(str(ans), str(ref))

        ans = self.walk(walker, m.x)
        ref = []
        self.assertEqual(str(ans), str(ref))

        ans = self.walk(walker, 2)
        ref = []
        self.assertEqual(str(ans), str(ref))

    def test_reduce_in_accept(self):
        def enter(node):
            return None, 1

        def accept(node, data, child_result, child_idx):
            return data + child_result

        walker = StreamBasedExpressionVisitor(enterNode=enter, acceptChildResult=accept)
        # 4 operators, 6 leaf nodes
        self.assertEqual(self.walk(walker, self.e), 10)

    def test_sizeof_expression(self):
        self.assertEqual(sizeof_expression(self.e), 10)

    def test_enterNode(self):
        # This is an alternative way to implement the beforeChild test:
        def enter(node):
            if type(node) in nonpyomo_leaf_types or not node.is_expression_type():
                return (), [node]
            return node.args, []

        walker = StreamBasedExpressionVisitor(enterNode=enter)
        m = self.m

        ans = self.walk(walker, self.e)
        ref = [[[m.x], [2]], [m.y], [[m.z], [[m.x], [m.y]]]]
        self.assertEqual(str(ans), str(ref))

        ans = self.walk(walker, m.x)
        ref = [m.x]
        self.assertEqual(str(ans), str(ref))

        ans = self.walk(walker, 2)
        ref = [2]
        self.assertEqual(str(ans), str(ref))

    def test_enterNode_noLeafList(self):
        # This is an alternative way to implement the beforeChild test:
        def enter(node):
            if type(node) in nonpyomo_leaf_types or not node.is_expression_type():
                return (), node
            return node.args, []

        walker = StreamBasedExpressionVisitor(enterNode=enter)
        m = self.m

        ans = self.walk(walker, self.e)
        ref = [[m.x, 2], m.y, [m.z, [m.x, m.y]]]
        self.assertEqual(str(ans), str(ref))

        ans = self.walk(walker, m.x)
        ref = m.x
        self.assertEqual(str(ans), str(ref))

        ans = self.walk(walker, 2)
        ref = 2
        self.assertEqual(str(ans), str(ref))

    def test_enterNode_withFinalize(self):
        # This is an alternative way to implement the beforeChild test:
        def enter(node):
            if type(node) in nonpyomo_leaf_types or not node.is_expression_type():
                return (), node
            return node.args, []

        def finalize(result):
            if type(result) is list:
                return result
            else:
                return [result]

        walker = StreamBasedExpressionVisitor(enterNode=enter, finalizeResult=finalize)
        m = self.m

        ans = self.walk(walker, self.e)
        ref = [[m.x, 2], m.y, [m.z, [m.x, m.y]]]
        self.assertEqual(str(ans), str(ref))

        ans = self.walk(walker, m.x)
        ref = [m.x]
        self.assertEqual(str(ans), str(ref))

        ans = self.walk(walker, 2)
        ref = [2]
        self.assertEqual(str(ans), str(ref))

    def test_exitNode(self):
        # This is an alternative way to implement the beforeChild test:
        def exit(node, data):
            if data:
                return data
            else:
                return [node]

        walker = StreamBasedExpressionVisitor(exitNode=exit)
        m = self.m

        ans = self.walk(walker, self.e)
        ref = [[[m.x], [2]], [m.y], [[m.z], [[m.x], [m.y]]]]
        self.assertEqual(str(ans), str(ref))

        ans = self.walk(walker, m.x)
        ref = [m.x]
        self.assertEqual(str(ans), str(ref))

        ans = self.walk(walker, 2)
        ref = [2]
        self.assertEqual(str(ans), str(ref))

    def test_beforeChild_acceptChildResult_afterChild(self):
        counts = [0, 0, 0]

        def before(node, child, child_idx):
            counts[0] += 1
            if type(child) in nonpyomo_leaf_types or not child.is_expression_type():
                return False, None

        def accept(node, data, child_result, child_idx):
            counts[1] += 1

        def after(node, child, child_idx):
            counts[2] += 1

        walker = StreamBasedExpressionVisitor(
            beforeChild=before, acceptChildResult=accept, afterChild=after
        )
        ans = self.walk(walker, self.e)
        m = self.m
        self.assertEqual(ans, None)
        self.assertEqual(counts, [9, 9, 9])

    def test_OLD_beforeChild_acceptChildResult_afterChild(self):
        counts = [0, 0, 0]

        def before(node, child):
            counts[0] += 1
            if type(child) in nonpyomo_leaf_types or not child.is_expression_type():
                return False, None

        def accept(node, data, child_result):
            counts[1] += 1

        def after(node, child):
            counts[2] += 1

        os = StringIO()
        with LoggingIntercept(os, 'pyomo'):
            walker = StreamBasedExpressionVisitor(
                beforeChild=before, acceptChildResult=accept, afterChild=after
            )
        self.assertIn(
            "Note that the API for the StreamBasedExpressionVisitor "
            "has changed to include the child index for the "
            "beforeChild() method",
            os.getvalue().replace('\n', ' '),
        )
        self.assertIn(
            "Note that the API for the StreamBasedExpressionVisitor "
            "has changed to include the child index for the "
            "acceptChildResult() method",
            os.getvalue().replace('\n', ' '),
        )
        self.assertIn(
            "Note that the API for the StreamBasedExpressionVisitor "
            "has changed to include the child index for the "
            "afterChild() method",
            os.getvalue().replace('\n', ' '),
        )

        ans = self.walk(walker, self.e)
        m = self.m
        self.assertEqual(ans, None)
        self.assertEqual(counts, [9, 9, 9])

    def test_enterNode_acceptChildResult_beforeChild(self):
        ans = []

        def before(node, child, child_idx):
            if type(child) in nonpyomo_leaf_types or not child.is_expression_type():
                return False, child

        def accept(node, data, child_result, child_idx):
            if data is not child_result:
                data.append(child_result)
            return data

        def enter(node):
            return node.args, ans

        walker = StreamBasedExpressionVisitor(
            enterNode=enter, beforeChild=before, acceptChildResult=accept
        )
        ans = self.walk(walker, self.e)
        m = self.m
        ref = [m.x, 2, m.y, m.z, m.x, m.y]
        self.assertEqual(str(ans), str(ref))

    def test_finalize(self):
        ans = []

        def before(node, child, child_idx):
            if type(child) in nonpyomo_leaf_types or not child.is_expression_type():
                return False, child

        def accept(node, data, child_result, child_idx):
            if data is not child_result:
                data.append(child_result)
            return data

        def enter(node):
            return node.args, ans

        def finalize(result):
            return len(result)

        walker = StreamBasedExpressionVisitor(
            enterNode=enter,
            beforeChild=before,
            acceptChildResult=accept,
            finalizeResult=finalize,
        )
        ans = self.walk(walker, self.e)
        self.assertEqual(ans, 6)

    def test_all_function_pointers(self):
        ans = []

        def name(x):
            if type(x) in nonpyomo_leaf_types:
                return str(x)
            else:
                return x.name

        def initialize(expr):
            ans.append("Initialize")
            return True, None

        def enter(node):
            ans.append("Enter %s" % (name(node)))

        def exit(node, data):
            ans.append("Exit %s" % (name(node)))

        def before(node, child, child_idx):
            ans.append("Before %s (from %s)" % (name(child), name(node)))

        def accept(node, data, child_result, child_idx):
            ans.append("Accept into %s" % (name(node)))

        def after(node, child, child_idx):
            ans.append("After %s (from %s)" % (name(child), name(node)))

        def finalize(result):
            ans.append("Finalize")

        walker = StreamBasedExpressionVisitor(
            initializeWalker=initialize,
            enterNode=enter,
            exitNode=exit,
            beforeChild=before,
            acceptChildResult=accept,
            afterChild=after,
            finalizeResult=finalize,
        )
        self.assertIsNone(self.walk(walker, self.e))
        self.assertEqual(
            "\n".join(ans),
            """Initialize
Enter sum
Before pow (from sum)
Enter pow
Before x (from pow)
Enter x
Exit x
Accept into pow
After x (from pow)
Before 2 (from pow)
Enter 2
Exit 2
Accept into pow
After 2 (from pow)
Exit pow
Accept into sum
After pow (from sum)
Before y (from sum)
Enter y
Exit y
Accept into sum
After y (from sum)
Before prod (from sum)
Enter prod
Before z (from prod)
Enter z
Exit z
Accept into prod
After z (from prod)
Before sum (from prod)
Enter sum
Before x (from sum)
Enter x
Exit x
Accept into sum
After x (from sum)
Before y (from sum)
Enter y
Exit y
Accept into sum
After y (from sum)
Exit sum
Accept into prod
After sum (from prod)
Exit prod
Accept into sum
After prod (from sum)
Exit sum
Finalize""",
        )

    def test_all_derived_class(self):
        def name(x):
            if type(x) in nonpyomo_leaf_types:
                return str(x)
            else:
                return x.name

        class all_callbacks(StreamBasedExpressionVisitor):
            def __init__(self):
                self.ans = []
                super(all_callbacks, self).__init__()

            def initializeWalker(self, expr):
                self.ans.append("Initialize")
                return True, None

            def enterNode(self, node):
                self.ans.append("Enter %s" % (name(node)))

            def exitNode(self, node, data):
                self.ans.append("Exit %s" % (name(node)))

            def beforeChild(self, node, child, child_idx):
                self.ans.append("Before %s (from %s)" % (name(child), name(node)))

            def acceptChildResult(self, node, data, child_result, child_idx):
                self.ans.append("Accept into %s" % (name(node)))

            def afterChild(self, node, child, child_idx):
                self.ans.append("After %s (from %s)" % (name(child), name(node)))

            def finalizeResult(self, result):
                self.ans.append("Finalize")

        walker = all_callbacks()
        self.assertIsNone(self.walk(walker, self.e))
        self.assertEqual(
            "\n".join(walker.ans),
            """Initialize
Enter sum
Before pow (from sum)
Enter pow
Before x (from pow)
Enter x
Exit x
Accept into pow
After x (from pow)
Before 2 (from pow)
Enter 2
Exit 2
Accept into pow
After 2 (from pow)
Exit pow
Accept into sum
After pow (from sum)
Before y (from sum)
Enter y
Exit y
Accept into sum
After y (from sum)
Before prod (from sum)
Enter prod
Before z (from prod)
Enter z
Exit z
Accept into prod
After z (from prod)
Before sum (from prod)
Enter sum
Before x (from sum)
Enter x
Exit x
Accept into sum
After x (from sum)
Before y (from sum)
Enter y
Exit y
Accept into sum
After y (from sum)
Exit sum
Accept into prod
After sum (from prod)
Exit prod
Accept into sum
After prod (from sum)
Exit sum
Finalize""",
        )

    def test_all_derived_class_oldAPI(self):
        def name(x):
            if type(x) in nonpyomo_leaf_types:
                return str(x)
            else:
                return x.name

        class all_callbacks(StreamBasedExpressionVisitor):
            def __init__(self):
                self.ans = []
                super(all_callbacks, self).__init__()

            def enterNode(self, node):
                self.ans.append("Enter %s" % (name(node)))

            def exitNode(self, node, data):
                self.ans.append("Exit %s" % (name(node)))

            def beforeChild(self, node, child):
                self.ans.append("Before %s (from %s)" % (name(child), name(node)))

            def acceptChildResult(self, node, data, child_result):
                self.ans.append("Accept into %s" % (name(node)))

            def afterChild(self, node, child):
                self.ans.append("After %s (from %s)" % (name(child), name(node)))

            def finalizeResult(self, result):
                self.ans.append("Finalize")

        os = StringIO()
        with LoggingIntercept(os, 'pyomo'):
            walker = all_callbacks()
        self.assertIn(
            "Note that the API for the StreamBasedExpressionVisitor "
            "has changed to include the child index for the "
            "beforeChild() method",
            os.getvalue().replace('\n', ' '),
        )
        self.assertIn(
            "Note that the API for the StreamBasedExpressionVisitor "
            "has changed to include the child index for the "
            "acceptChildResult() method",
            os.getvalue().replace('\n', ' '),
        )
        self.assertIn(
            "Note that the API for the StreamBasedExpressionVisitor "
            "has changed to include the child index for the "
            "afterChild() method",
            os.getvalue().replace('\n', ' '),
        )

        self.assertIsNone(self.walk(walker, self.e))
        self.assertEqual(
            "\n".join(walker.ans),
            """Enter sum
Before pow (from sum)
Enter pow
Before x (from pow)
Enter x
Exit x
Accept into pow
After x (from pow)
Before 2 (from pow)
Enter 2
Exit 2
Accept into pow
After 2 (from pow)
Exit pow
Accept into sum
After pow (from sum)
Before y (from sum)
Enter y
Exit y
Accept into sum
After y (from sum)
Before prod (from sum)
Enter prod
Before z (from prod)
Enter z
Exit z
Accept into prod
After z (from prod)
Before sum (from prod)
Enter sum
Before x (from sum)
Enter x
Exit x
Accept into sum
After x (from sum)
Before y (from sum)
Enter y
Exit y
Accept into sum
After y (from sum)
Exit sum
Accept into prod
After sum (from prod)
Exit prod
Accept into sum
After prod (from sum)
Exit sum
Finalize""",
        )


class TestStreamBasedExpressionVisitor_Recursive(
    BaseStreamBasedVisitorTests, unittest.TestCase
):
    def walk(self, walker, expr):
        return walker.walk_expression(expr)


class TestStreamBasedExpressionVisitor_NonRecursive(
    BaseStreamBasedVisitorTests, unittest.TestCase
):
    def walk(self, walker, expr):
        return walker.walk_expression_nonrecursive(expr)


def fill_stack(n, fcn, *args):
    if n:
        return fill_stack(n - 1, fcn, *args)
    else:
        return fcn(*args)


class TestStreamBasedExpressionVisitor_Deep(unittest.TestCase):
    def setUp(self):
        self.m = m = ConcreteModel()
        m.x = Var()
        m.I = Set(initialize=range(2 * RECURSION_LIMIT))

        def _rule(m, i):
            if i:
                return m.e[i - 1]
            else:
                return m.x

        m.e = Expression(m.I, rule=_rule)

    def evaluate_bx(self):
        def before(node, child, idx):
            if type(child) in native_types or not child.is_expression_type():
                return False, value(child)
            return True, None

        def exit(node, data):
            return data[0] + 1

        return StreamBasedExpressionVisitor(beforeChild=before, exitNode=exit)

    def evaluate_bex(self):
        def before(node, child, idx):
            if type(child) in native_types or not child.is_expression_type():
                return False, value(child)
            return True, None

        def enter(node):
            return None, []

        def exit(node, data):
            return data[0] + 1

        return StreamBasedExpressionVisitor(
            beforeChild=before, enterNode=enter, exitNode=exit
        )

    def evaluate_abex(self):
        def before(node, child, idx):
            if type(child) in native_types or not child.is_expression_type():
                return False, value(child)
            return True, None

        def enter(node):
            return None, 0

        def accept(node, data, child_result, child_idx):
            return data + child_result

        def exit(node, data):
            return data + 1

        return StreamBasedExpressionVisitor(
            beforeChild=before, acceptChildResult=accept, enterNode=enter, exitNode=exit
        )

    def run_walker(self, walker):
        m = self.m
        m.x = 10
        self.assertEqual(
            2 * RECURSION_LIMIT + 10,
            walker.walk_expression(m.e[2 * RECURSION_LIMIT - 1]),
        )
        self.assertEqual(
            2 * RECURSION_LIMIT + 10,
            walker.walk_expression_nonrecursive(m.e[2 * RECURSION_LIMIT - 1]),
        )

        # This is a "magic parameter" that quantifies the overhead
        # needed by the system to convert the recursive walker to a
        # nonrecursive one.
        #
        # Note: this needs to be 13 if pytest is run as a script, and 14
        # if pytest is run as "python -m".  We will use 14, and then add
        # 2 (instead of 1) to generate the recursion error.  Note that
        # the stack handling is different on GHA, and we need to fill an
        # additional frame (for a total of 3) to trigger the recursion
        # error.
        #
        TESTING_OVERHEAD = 14
        warn_msg = "Unexpected RecursionError walking an expression tree.\n"

        if platform.python_implementation() == 'PyPy':
            # We have not yet determined how to trigger the
            # RecursionError on PyPy
            cases = [(0, "")]
        elif os.environ.get('GITHUB_ACTIONS', '') and sys.platform.startswith('win'):
            # The test for handling RecursionError appears to fail
            # inexplicably on GHA/Windows under pytest: the
            # RecursionError that is supposed to be raised is not
            # raised, and instead the system actually dies on stack
            # overflow error
            cases = []
        else:
            # 3 sufficed through Python 3.10, but appeared to need to be
            # raised to 5 for Python 3.11 builds (3.11.2), and again to
            # 10 for Python 3.12 builds (3.12.0)
            cases = [(0, ""), (10, warn_msg)]

        head_room = sys.getrecursionlimit() - get_stack_depth()
        for n, msg in cases:
            with LoggingIntercept() as LOG:
                self.assertEqual(
                    2 * RECURSION_LIMIT + 10,
                    fill_stack(
                        head_room - RECURSION_LIMIT - TESTING_OVERHEAD + n,
                        walker.walk_expression,
                        m.e[2 * RECURSION_LIMIT - 1],
                    ),
                )
            self.assertEqual(msg, LOG.getvalue())

    def test_evaluate_bx(self):
        return self.run_walker(self.evaluate_bx())

    def test_evaluate_bex(self):
        return self.run_walker(self.evaluate_bex())

    def test_evaluate_abex(self):
        return self.run_walker(self.evaluate_abex())


class TestSimpleExpressionVisitor(unittest.TestCase):
    def test_base_class(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        m.p = Param(mutable=True)
        v = SimpleExpressionVisitor()

        e = 5
        self.assertEqual(v.xbfs(e), None)
        self.assertEqual(list(v.xbfs_yield_leaves(e)), [])

        e = m.x
        self.assertEqual(v.xbfs(e), None)
        self.assertEqual(list(v.xbfs_yield_leaves(e)), [])

        e = m.x + 5 * m.y**m.p
        self.assertEqual(v.xbfs(e), None)
        self.assertEqual(list(v.xbfs_yield_leaves(e)), [])

    def test_derived_visitor(self):
        class _Visitor(SimpleExpressionVisitor):
            def __init__(self):
                super().__init__()
                self.nodes = []

            def visit(self, node):
                self.nodes.append(node)
                return node

            def finalize(self):
                return len(self.nodes)

        m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        m.p = Param(mutable=True)
        v = _Visitor()

        e = 5
        self.assertEqual(v.xbfs(e), 1)
        self.assertEqual(list(v.xbfs_yield_leaves(e)), [5])

        e = m.x
        self.assertEqual(v.xbfs(e), 3)
        self.assertEqual(list(v.xbfs_yield_leaves(e)), [m.x])

        e = m.x + 5 * m.y**m.p
        self.assertEqual(v.xbfs(e), 11)
        self.assertEqual(list(v.xbfs_yield_leaves(e)), [m.x, 5, m.y, m.p])


class TestEvaluateExpression(unittest.TestCase):
    def test_constant(self):
        m = ConcreteModel()
        m.p = Param(initialize=1)

        e = 1 + m.p
        self.assertEqual(2, evaluate_expression(e))
        self.assertEqual(2, evaluate_expression(e, constant=True))

    def test_uninitialized_constant(self):
        m = ConcreteModel()
        m.p = Param(mutable=True)

        e = 1 + m.p
        self.assertRaises(ValueError, evaluate_expression, e)
        self.assertRaises(FixedExpressionError, evaluate_expression, e, constant=True)

    def test_variable(self):
        m = ConcreteModel()
        m.p = Var()

        e = 1 + m.p
        self.assertRaises(ValueError, evaluate_expression, e)
        self.assertRaises(
            NonConstantExpressionError, evaluate_expression, e, constant=True
        )

    def test_initialized_variable(self):
        m = ConcreteModel()
        m.p = Var(initialize=1)

        e = 1 + m.p
        self.assertTrue(2, evaluate_expression(e))
        self.assertRaises(
            NonConstantExpressionError, evaluate_expression, e, constant=True
        )

    def test_fixed_variable(self):
        m = ConcreteModel()
        m.p = Var(initialize=1)
        m.p.fixed = True

        e = 1 + m.p
        self.assertTrue(2, evaluate_expression(e))
        self.assertRaises(FixedExpressionError, evaluate_expression, e, constant=True)

    def test_template_expr(self):
        m = ConcreteModel()
        m.I = RangeSet(1, 9)
        m.x = Var(m.I, initialize=lambda m, i: i + 1)
        m.P = Param(m.I, initialize=lambda m, i: 10 - i, mutable=True)
        t = IndexTemplate(m.I)

        e = m.x[t + m.P[t + 1]] + 3
        self.assertRaises(TemplateExpressionError, evaluate_expression, e)
        self.assertRaises(
            TemplateExpressionError, evaluate_expression, e, constant=True
        )


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.incidence_analysis.matching import maximum_matching
from pyomo.common.dependencies import scipy, scipy_available, networkx_available

if scipy_available:
    sps = scipy.sparse

import pyomo.common.unittest as unittest


@unittest.skipUnless(networkx_available, "networkx is not available")
@unittest.skipUnless(scipy_available, "scipy is not available")
class TestMatching(unittest.TestCase):
    def test_identity(self):
        N = 5
        matrix = sps.identity(N)
        matching = maximum_matching(matrix)
        self.assertEqual(len(matching), N)
        for i in range(N):
            self.assertIn(i, matching)
            self.assertEqual(i, matching[i])

        matrix = matrix.tocoo()
        matching = maximum_matching(matrix)
        self.assertEqual(len(matching), N)
        for i in range(N):
            self.assertIn(i, matching)
            self.assertEqual(i, matching[i])

        matrix = matrix.tocsc()
        matching = maximum_matching(matrix)
        self.assertEqual(len(matching), N)
        for i in range(N):
            self.assertIn(i, matching)
            self.assertEqual(i, matching[i])

    def test_low_rank_diagonal(self):
        N = 5
        omit = N // 2
        row = [i for i in range(N) if i != omit]
        col = [j for j in range(N) if j != omit]
        data = [1 for _ in range(N - 1)]
        matrix = sps.coo_matrix((data, (row, col)), shape=(N, N))
        matching = maximum_matching(matrix)

        self.assertEqual(len(matching), N - 1)
        for i in range(N):
            if i != omit:
                self.assertIn(i, matching)
                self.assertEqual(i, matching[i])

    def test_bordered(self):
        N = 5
        row = []
        col = []
        data = []
        for i in range(N - 1):
            # Bottom row
            row.append(N - 1)
            col.append(i)
            data.append(1)

            # Right column
            row.append(i)
            col.append(N - 1)
            data.append(1)

            # Diagonal
            row.append(i)
            col.append(i)
            data.append(1)

        matrix = sps.coo_matrix((data, (row, col)), shape=(N, N))
        matching = maximum_matching(matrix)

        self.assertEqual(len(matching), N)
        values = set(matching.values())
        for i in range(N):
            self.assertIn(i, matching)
            self.assertIn(i, values)

    def test_hessenberg(self):
        """
        |x x      |
        |    x    |
        |      x  |
        |        x|
        |x x x x x|
        """
        N = 5
        row = []
        col = []
        data = []
        for i in range(N):
            # Bottom row
            row.append(N - 1)
            col.append(i)
            data.append(1)

            if i == 0:
                # Top left entry
                row.append(0)
                col.append(i)
                data.append(1)
            else:
                # One-off diagonal
                row.append(i - 1)
                col.append(i)
                data.append(1)

        matrix = sps.coo_matrix((data, (row, col)), shape=(N, N))
        matching = maximum_matching(matrix)

        self.assertEqual(len(matching), N)
        values = set(matching.values())
        for i in range(N):
            self.assertIn(i, matching)
            self.assertIn(i, values)

    def test_low_rank_hessenberg(self):
        """
        |x x      |
        |         |
        |      x  |
        |        x|
        |x x x x x|
        Know that first and last row and column will be in
        the imperfect matching.
        """
        N = 5
        omit = N // 2
        row = []
        col = []
        data = []
        for i in range(N):
            # Bottom row
            row.append(N - 1)
            col.append(i)
            data.append(1)

            if i == 0:
                # Top left entry
                row.append(0)
                col.append(i)
                data.append(1)
            else:
                # One-off diagonal
                if i != omit:
                    row.append(i - 1)
                    col.append(i)
                    data.append(1)

        matrix = sps.coo_matrix((data, (row, col)), shape=(N, N))
        matching = maximum_matching(matrix)
        values = set(matching.values())

        self.assertEqual(len(matching), N - 1)
        self.assertIn(0, matching)
        self.assertIn(N - 1, matching)
        self.assertIn(0, values)
        self.assertIn(N - 1, values)

    def test_nondecomposable_hessenberg(self):
        """
        |x x      |
        |  x x    |
        |    x x  |
        |      x x|
        |x x x x x|
        """
        N = 5
        row = []
        col = []
        data = []
        for i in range(N):
            # Bottom row
            row.append(N - 1)
            col.append(i)
            data.append(1)

            # Diagonal
            row.append(i)
            col.append(i)
            data.append(1)
            # ^ This will put another entry at (N-1, N-1).
            # This is fine.

            if i != 0:
                # One-off diagonal
                row.append(i - 1)
                col.append(i)
                data.append(1)

        matrix = sps.coo_matrix((data, (row, col)), shape=(N, N))
        matching = maximum_matching(matrix)
        values = set(matching.values())

        self.assertEqual(len(matching), N)
        for i in range(N):
            self.assertIn(i, matching)
            self.assertIn(i, values)

    def test_low_rank_nondecomposable_hessenberg(self):
        """
        |  x      |
        |x   x    |
        |  x   x  |
        |    x   x|
        |      x  |
        """
        N = 5
        # For N odd, a matrix with this structure does not have
        # a perfect matching.
        row = []
        col = []
        data = []
        for i in range(N - 1):
            # Below diagonal
            row.append(i + 1)
            col.append(i)
            data.append(1)

            # Above diagonal
            row.append(i)
            col.append(i + 1)
            data.append(1)

        matrix = sps.coo_matrix((data, (row, col)), shape=(N, N))
        matching = maximum_matching(matrix)
        values = set(matching.values())

        self.assertEqual(len(matching), N - 1)
        self.assertEqual(len(values), N - 1)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os.path
from pyomo.common.dependencies import attempt_import
from pyomo.dataportal.factory import DataManagerFactory
from pyomo.dataportal import TableData


def _xml_importer():
    try:
        from lxml import etree

        return etree
    except ImportError:
        pass

    try:
        # Python 2.5+
        import xml.etree.cElementTree as etree

        return etree
    except ImportError:
        pass

    # Python 2.5+
    import xml.etree.ElementTree as etree

    return etree


ET, ET_available = attempt_import('ET', importer=_xml_importer)


@DataManagerFactory.register("xml", "XML file interface")
class XMLTable(TableData):
    def __init__(self):
        TableData.__init__(self)

    def open(self):
        if self.filename is None:
            raise IOError("No filename specified")

    def close(self):
        pass

    def read(self):
        from pyomo.core.base.param import Param

        if not os.path.exists(self.filename):
            raise IOError("Cannot find file '%s'" % self.filename)
        #
        tree = ET.parse(self.filename)
        if not self.options.query is None:
            if self.options.query[0] == '"' or self.options.query[0] == "'":
                self.options.query = self.options.query[1:-1]
            parents = [parent for parent in tree.findall(self.options.query)]
        else:
            parents = [parent for parent in tree.getroot()]
        tmp = []
        labels = []
        for parent in parents:
            if len(tmp) == 0:
                for child in parent:
                    labels.append(child.tag)
                tmp.append(labels)
            row = {}
            for child in parent:
                if child.text is None:
                    row[child.tag] = child.get('value')
                else:
                    row[child.tag] = child.text
            tmp.append([row.get(label, '.') for label in labels])
        #
        if len(tmp) == 0:
            raise IOError("Empty *.xml file")
        elif len(tmp) == 1:
            if not self.options.param is None:
                if type(self.options.param) in (list, tuple):
                    p = self.options.param[0]
                else:
                    p = self.options.param
                if isinstance(p, Param):
                    self.options.model = p._model()
                    p = p.local_name
                self._info = ["param", p, ":=", tmp[0][0]]
            elif len(self.options.symbol_map) == 1:
                self._info = [
                    "param",
                    self.options.symbol_map[self.options.symbol_map.keys()[0]],
                    ":=",
                    tmp[0][0],
                ]
            else:
                raise IOError(
                    "Data looks like a parameter, but multiple parameter names have been specified: %s"
                    % str(self.options.symbol_map)
                )
        else:
            self._set_data(tmp[0], tmp[1:])

    def write(self, data):
        if self.options.set is None and self.options.param is None:
            raise IOError("Unspecified model component")
        root = ET.Element('table')
        table = self._get_table()
        labels = table[0]
        for i in range(len(labels)):
            labels[i] = str(labels[i])
        for trow in table[1:]:
            row = ET.SubElement(root, 'row')
            for i in range(len(labels)):
                data = ET.SubElement(row, labels[i])
                data.set('value', str(trow[i]))
        #
        tree = ET.ElementTree(root)
        tree.write(self.filename)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.pynumero.dependencies import numpy_available, scipy_available
import pyomo.common.unittest as unittest

if not (numpy_available and scipy_available):
    raise unittest.SkipTest('Pynumero examples need scipy and numpy')

import numpy as np

from pyomo.contrib.pynumero.asl import AmplInterface

if not AmplInterface.available():
    raise unittest.SkipTest('Pynumero examples need ASL')

from pyomo.contrib.pynumero.linalg.mumps_interface import mumps_available
from pyomo.contrib.pynumero.linalg.scipy_interface import ScipyLU

import pyomo.environ as pyo

ipopt_opt = pyo.SolverFactory('ipopt')
ipopt_available = ipopt_opt.available(exception_flag=False)

from pyomo.contrib.pynumero.examples import (
    nlp_interface,
    nlp_interface_2,
    feasibility,
    mumps_example,
    sensitivity,
    sqp,
)


class TestPyNumeroExamples(unittest.TestCase):
    def test_nlp_interface(self):
        nlp_interface.main()

    def test_nlp_interface_2(self):
        nlp_interface_2.main(show_plot=False)

    @unittest.skipIf(not ipopt_available, "feasibility example requires ipopt")
    def test_feasibility(self):
        is_feasible = feasibility.main()
        self.assertTrue(is_feasible)

    @unittest.skipIf(not mumps_available, 'mumps example needs pymumps')
    def test_mumps_example(self):
        mumps_example.main()

    @unittest.skipIf(not ipopt_available, "sensitivity example requires ipopt")
    def test_sensitivity(self):
        x_sens, x_correct = sensitivity.main()
        self.assertTrue(np.allclose(x_sens, x_correct, rtol=1e-3, atol=1e-4))

    def test_sqp(self):
        obj = sqp.main(ScipyLU(), 10, 10)
        self.assertAlmostEqual(obj, 4.9834689888961198e-02)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import sys

from pyomo.common.deprecation import relocated_module_attribute

logger = logging.getLogger('pyomo.core')

relocated_module_attribute(
    'tabular_writer', 'pyomo.common.formatting.tabular_writer', version='6.1'
)
relocated_module_attribute(
    'sorted_robust', 'pyomo.common.sorting.sorted_robust', version='6.1'
)


def display(obj, ostream=None):
    """Display data in a Pyomo object"""
    if ostream is None:
        ostream = sys.stdout
    try:
        display_fcn = obj.display
    except AttributeError:
        raise TypeError(
            "Error trying to display values for object of type %s:\n"
            "\tObject does not support the 'display()' method" % (type(obj),)
        )
    try:
        display_fcn(ostream=ostream)
    except Exception:
        err = sys.exc_info()[1]
        logger.error(
            "Error trying to display values for object of type %s:\n\t%s"
            % (type(obj), err)
        )
        raise


def create_name(name, ndx):
    """Create a canonical name for a component using the given index"""
    if ndx is None:
        return name
    if type(ndx) is tuple:
        tmp = str(ndx).replace(', ', ',')
        return name + "[" + tmp[1:-1] + "]"
    return name + "[" + str(ndx) + "]"


def apply_indexed_rule(obj, rule, model, index, options=None):
    try:
        if options is None:
            if index.__class__ is tuple:
                return rule(model, *index)
            elif index is None and not obj.is_indexed():
                return rule(model)
            else:
                return rule(model, index)
        else:
            if index.__class__ is tuple:
                return rule(model, *index, **options)
            elif index is None and not obj.is_indexed():
                return rule(model, **options)
            else:
                return rule(model, index, **options)
    except TypeError:
        try:
            if options is None:
                return rule(model)
            else:
                return rule(model, **options)
        except:
            # Nothing appears to have matched... re-trigger the original
            # TypeError
            if options is None:
                if index.__class__ is tuple:
                    return rule(model, *index)
                elif index is None and not obj.is_indexed():
                    return rule(model)
                else:
                    return rule(model, index)
            else:
                if index.__class__ is tuple:
                    return rule(model, *index, **options)
                elif index is None and not obj.is_indexed():
                    return rule(model, **options)
                else:
                    return rule(model, index, **options)


def apply_parameterized_indexed_rule(obj, rule, model, param, index):
    if index.__class__ is tuple:
        return rule(model, param, *index)
    if index is None:
        return rule(model, param)
    return rule(model, param, index)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import abc
import datetime
import time
from typing import List

from pyomo.core.base.constraint import ConstraintData, Constraint
from pyomo.core.base.sos import SOSConstraintData, SOSConstraint
from pyomo.core.base.var import VarData
from pyomo.core.base.param import ParamData, Param
from pyomo.core.base.objective import ObjectiveData
from pyomo.core.staleflag import StaleFlagManager
from pyomo.common.collections import ComponentMap
from pyomo.common.timing import HierarchicalTimer
from pyomo.contrib.solver.common.results import Results
from pyomo.contrib.solver.common.util import collect_vars_and_named_exprs, get_objective


class PersistentSolverUtils(abc.ABC):
    def __init__(self, treat_fixed_vars_as_params=True):
        """
        Parameters
        ----------
        treat_fixed_vars_as_params: bool
            This is an advanced option that should only be used in special circumstances.
            With the default setting of True, fixed variables will be treated like parameters.
            This means that z == x*y will be linear if x or y is fixed and the constraint
            can be written to an LP file. If the value of the fixed variable gets changed, we have
            to completely reprocess all constraints using that variable. If
            treat_fixed_vars_as_params is False, then constraints will be processed as if fixed
            variables are not fixed, and the solver will be told the variable is fixed. This means
            z == x*y could not be written to an LP file even if x and/or y is fixed. However,
            updating the values of fixed variables is much faster this way.
        """
        self._model = None
        self._active_constraints = {}  # maps constraint to (lower, body, upper)
        self._vars = {}  # maps var id to (var, lb, ub, fixed, domain, value)
        self._params = {}  # maps param id to param
        self._objective = None
        self._objective_expr = None
        self._objective_sense = None
        self._named_expressions = (
            {}
        )  # maps constraint to list of tuples (named_expr, named_expr.expr)
        self._external_functions = ComponentMap()
        self._obj_named_expressions = []
        self._referenced_variables = (
            {}
        )  # var_id: [dict[constraints, None], dict[sos constraints, None], None or objective]
        self._vars_referenced_by_con = {}
        self._vars_referenced_by_obj = []
        self._expr_types = None
        self._treat_fixed_vars_as_params = treat_fixed_vars_as_params
        self._active_config = self.config

    def set_instance(self, model):
        saved_config = self.config
        saved_active_config = self._active_config
        self.__init__()
        self.config = saved_config
        self._active_config = saved_active_config
        self._model = model
        self.add_block(model)
        if self._objective is None:
            self.set_objective(None)

    @abc.abstractmethod
    def _add_variables(self, variables: List[VarData]):
        pass

    def add_variables(self, variables: List[VarData]):
        for v in variables:
            if id(v) in self._referenced_variables:
                raise ValueError(f'Variable {v.name} has already been added')
            self._referenced_variables[id(v)] = [{}, {}, None]
            self._vars[id(v)] = (
                v,
                v._lb,
                v._ub,
                v.fixed,
                v.domain.get_interval(),
                v.value,
            )
        self._add_variables(variables)

    @abc.abstractmethod
    def _add_parameters(self, params: List[ParamData]):
        pass

    def add_parameters(self, params: List[ParamData]):
        for p in params:
            self._params[id(p)] = p
        self._add_parameters(params)

    @abc.abstractmethod
    def _add_constraints(self, cons: List[ConstraintData]):
        pass

    def _check_for_new_vars(self, variables: List[VarData]):
        new_vars = {}
        for v in variables:
            v_id = id(v)
            if v_id not in self._referenced_variables:
                new_vars[v_id] = v
        self.add_variables(list(new_vars.values()))

    def _check_to_remove_vars(self, variables: List[VarData]):
        vars_to_remove = {}
        for v in variables:
            v_id = id(v)
            ref_cons, ref_sos, ref_obj = self._referenced_variables[v_id]
            if len(ref_cons) == 0 and len(ref_sos) == 0 and ref_obj is None:
                vars_to_remove[v_id] = v
        self.remove_variables(list(vars_to_remove.values()))

    def add_constraints(self, cons: List[ConstraintData]):
        all_fixed_vars = {}
        for con in cons:
            if con in self._named_expressions:
                raise ValueError(f'Constraint {con.name} has already been added')
            self._active_constraints[con] = con.expr
            tmp = collect_vars_and_named_exprs(con.expr)
            named_exprs, variables, fixed_vars, external_functions = tmp
            self._check_for_new_vars(variables)
            self._named_expressions[con] = [(e, e.expr) for e in named_exprs]
            if len(external_functions) > 0:
                self._external_functions[con] = external_functions
            self._vars_referenced_by_con[con] = variables
            for v in variables:
                self._referenced_variables[id(v)][0][con] = None
            if not self._treat_fixed_vars_as_params:
                for v in fixed_vars:
                    v.unfix()
                    all_fixed_vars[id(v)] = v
        self._add_constraints(cons)
        for v in all_fixed_vars.values():
            v.fix()

    @abc.abstractmethod
    def _add_sos_constraints(self, cons: List[SOSConstraintData]):
        pass

    def add_sos_constraints(self, cons: List[SOSConstraintData]):
        for con in cons:
            if con in self._vars_referenced_by_con:
                raise ValueError(f'Constraint {con.name} has already been added')
            self._active_constraints[con] = tuple()
            variables = con.get_variables()
            self._check_for_new_vars(variables)
            self._named_expressions[con] = []
            self._vars_referenced_by_con[con] = variables
            for v in variables:
                self._referenced_variables[id(v)][1][con] = None
        self._add_sos_constraints(cons)

    @abc.abstractmethod
    def _set_objective(self, obj: ObjectiveData):
        pass

    def set_objective(self, obj: ObjectiveData):
        if self._objective is not None:
            for v in self._vars_referenced_by_obj:
                self._referenced_variables[id(v)][2] = None
            self._check_to_remove_vars(self._vars_referenced_by_obj)
            self._external_functions.pop(self._objective, None)
        if obj is not None:
            self._objective = obj
            self._objective_expr = obj.expr
            self._objective_sense = obj.sense
            tmp = collect_vars_and_named_exprs(obj.expr)
            named_exprs, variables, fixed_vars, external_functions = tmp
            self._check_for_new_vars(variables)
            self._obj_named_expressions = [(i, i.expr) for i in named_exprs]
            if len(external_functions) > 0:
                self._external_functions[obj] = external_functions
            self._vars_referenced_by_obj = variables
            for v in variables:
                self._referenced_variables[id(v)][2] = obj
            if not self._treat_fixed_vars_as_params:
                for v in fixed_vars:
                    v.unfix()
            self._set_objective(obj)
            for v in fixed_vars:
                v.fix()
        else:
            self._vars_referenced_by_obj = []
            self._objective = None
            self._objective_expr = None
            self._objective_sense = None
            self._obj_named_expressions = []
            self._set_objective(obj)

    def add_block(self, block):
        param_dict = {}
        for p in block.component_objects(Param, descend_into=True):
            if p.mutable:
                for _p in p.values():
                    param_dict[id(_p)] = _p
        self.add_parameters(list(param_dict.values()))
        self.add_constraints(
            list(
                block.component_data_objects(Constraint, descend_into=True, active=True)
            )
        )
        self.add_sos_constraints(
            list(
                block.component_data_objects(
                    SOSConstraint, descend_into=True, active=True
                )
            )
        )
        obj = get_objective(block)
        if obj is not None:
            self.set_objective(obj)

    @abc.abstractmethod
    def _remove_constraints(self, cons: List[ConstraintData]):
        pass

    def remove_constraints(self, cons: List[ConstraintData]):
        self._remove_constraints(cons)
        for con in cons:
            if con not in self._named_expressions:
                raise ValueError(
                    f'Cannot remove constraint {con.name} - it was not added'
                )
            for v in self._vars_referenced_by_con[con]:
                self._referenced_variables[id(v)][0].pop(con)
            self._check_to_remove_vars(self._vars_referenced_by_con[con])
            del self._active_constraints[con]
            del self._named_expressions[con]
            self._external_functions.pop(con, None)
            del self._vars_referenced_by_con[con]

    @abc.abstractmethod
    def _remove_sos_constraints(self, cons: List[SOSConstraintData]):
        pass

    def remove_sos_constraints(self, cons: List[SOSConstraintData]):
        self._remove_sos_constraints(cons)
        for con in cons:
            if con not in self._vars_referenced_by_con:
                raise ValueError(
                    f'Cannot remove constraint {con.name} - it was not added'
                )
            for v in self._vars_referenced_by_con[con]:
                self._referenced_variables[id(v)][1].pop(con)
            self._check_to_remove_vars(self._vars_referenced_by_con[con])
            del self._active_constraints[con]
            del self._named_expressions[con]
            del self._vars_referenced_by_con[con]

    @abc.abstractmethod
    def _remove_variables(self, variables: List[VarData]):
        pass

    def remove_variables(self, variables: List[VarData]):
        self._remove_variables(variables)
        for v in variables:
            v_id = id(v)
            if v_id not in self._referenced_variables:
                raise ValueError(
                    f'Cannot remove variable {v.name} - it has not been added'
                )
            cons_using, sos_using, obj_using = self._referenced_variables[v_id]
            if cons_using or sos_using or (obj_using is not None):
                raise ValueError(
                    f'Cannot remove variable {v.name} - it is still being used by constraints or the objective'
                )
            del self._referenced_variables[v_id]
            del self._vars[v_id]

    @abc.abstractmethod
    def _remove_parameters(self, params: List[ParamData]):
        pass

    def remove_parameters(self, params: List[ParamData]):
        self._remove_parameters(params)
        for p in params:
            del self._params[id(p)]

    def remove_block(self, block):
        self.remove_constraints(
            list(
                block.component_data_objects(
                    ctype=Constraint, descend_into=True, active=True
                )
            )
        )
        self.remove_sos_constraints(
            list(
                block.component_data_objects(
                    ctype=SOSConstraint, descend_into=True, active=True
                )
            )
        )
        self.remove_parameters(
            list(
                dict(
                    (id(p), p)
                    for p in block.component_data_objects(
                        ctype=Param, descend_into=True
                    )
                ).values()
            )
        )

    @abc.abstractmethod
    def _update_variables(self, variables: List[VarData]):
        pass

    def update_variables(self, variables: List[VarData]):
        for v in variables:
            self._vars[id(v)] = (
                v,
                v._lb,
                v._ub,
                v.fixed,
                v.domain.get_interval(),
                v.value,
            )
        self._update_variables(variables)

    @abc.abstractmethod
    def update_parameters(self):
        pass

    def update(self, timer: HierarchicalTimer = None):
        if timer is None:
            timer = HierarchicalTimer()
        config = self._active_config.auto_updates
        new_vars = []
        old_vars = []
        new_params = []
        old_params = []
        new_cons = []
        old_cons = []
        old_sos = []
        new_sos = []
        current_cons_dict = {}
        current_sos_dict = {}
        timer.start('vars')
        if config.update_vars:
            start_vars = {v_id: v_tuple[0] for v_id, v_tuple in self._vars.items()}
        timer.stop('vars')
        timer.start('params')
        if config.check_for_new_or_removed_params:
            current_params_dict = {}
            for p in self._model.component_objects(Param, descend_into=True):
                if p.mutable:
                    for _p in p.values():
                        current_params_dict[id(_p)] = _p
            for p_id, p in current_params_dict.items():
                if p_id not in self._params:
                    new_params.append(p)
            for p_id, p in self._params.items():
                if p_id not in current_params_dict:
                    old_params.append(p)
        timer.stop('params')
        timer.start('cons')
        if config.check_for_new_or_removed_constraints or config.update_constraints:
            current_cons_dict = {
                c: None
                for c in self._model.component_data_objects(
                    Constraint, descend_into=True, active=True
                )
            }
            current_sos_dict = {
                c: None
                for c in self._model.component_data_objects(
                    SOSConstraint, descend_into=True, active=True
                )
            }
            for c in current_cons_dict.keys():
                if c not in self._vars_referenced_by_con:
                    new_cons.append(c)
            for c in current_sos_dict.keys():
                if c not in self._vars_referenced_by_con:
                    new_sos.append(c)
            for c in self._vars_referenced_by_con:
                if c not in current_cons_dict and c not in current_sos_dict:
                    if (c.ctype is Constraint) or (
                        c.ctype is None and isinstance(c, ConstraintData)
                    ):
                        old_cons.append(c)
                    else:
                        assert (c.ctype is SOSConstraint) or (
                            c.ctype is None and isinstance(c, SOSConstraintData)
                        )
                        old_sos.append(c)
        self.remove_constraints(old_cons)
        self.remove_sos_constraints(old_sos)
        timer.stop('cons')
        timer.start('params')
        self.remove_parameters(old_params)

        # sticking this between removal and addition
        # is important so that we don't do unnecessary work
        if config.update_parameters:
            self.update_parameters()

        self.add_parameters(new_params)
        timer.stop('params')
        timer.start('vars')
        self.add_variables(new_vars)
        timer.stop('vars')
        timer.start('cons')
        self.add_constraints(new_cons)
        self.add_sos_constraints(new_sos)
        new_cons_set = set(new_cons)
        new_sos_set = set(new_sos)
        cons_to_remove_and_add = {}
        need_to_set_objective = False
        if config.update_constraints:
            for c in current_cons_dict.keys():
                if c not in new_cons_set and c.expr is not self._active_constraints[c]:
                    cons_to_remove_and_add[c] = None
            sos_to_update = []
            for c in current_sos_dict.keys():
                if c not in new_sos_set:
                    sos_to_update.append(c)
            self.remove_sos_constraints(sos_to_update)
            self.add_sos_constraints(sos_to_update)
        timer.stop('cons')
        timer.start('vars')
        if config.update_vars:
            end_vars = {v_id: v_tuple[0] for v_id, v_tuple in self._vars.items()}
            vars_to_check = [v for v_id, v in end_vars.items() if v_id in start_vars]
        if config.update_vars:
            vars_to_update = []
            for v in vars_to_check:
                _v, lb, ub, fixed, domain_interval, value = self._vars[id(v)]
                if (fixed != v.fixed) or (fixed and (value != v.value)):
                    vars_to_update.append(v)
                    if self._treat_fixed_vars_as_params:
                        for c in self._referenced_variables[id(v)][0]:
                            cons_to_remove_and_add[c] = None
                        if self._referenced_variables[id(v)][2] is not None:
                            need_to_set_objective = True
                elif lb is not v._lb:
                    vars_to_update.append(v)
                elif ub is not v._ub:
                    vars_to_update.append(v)
                elif domain_interval != v.domain.get_interval():
                    vars_to_update.append(v)
            self.update_variables(vars_to_update)
        timer.stop('vars')
        timer.start('cons')
        cons_to_remove_and_add = list(cons_to_remove_and_add.keys())
        self.remove_constraints(cons_to_remove_and_add)
        self.add_constraints(cons_to_remove_and_add)
        timer.stop('cons')
        timer.start('named expressions')
        if config.update_named_expressions:
            cons_to_update = []
            for c, expr_list in self._named_expressions.items():
                if c in new_cons_set:
                    continue
                for named_expr, old_expr in expr_list:
                    if named_expr.expr is not old_expr:
                        cons_to_update.append(c)
                        break
            self.remove_constraints(cons_to_update)
            self.add_constraints(cons_to_update)
            for named_expr, old_expr in self._obj_named_expressions:
                if named_expr.expr is not old_expr:
                    need_to_set_objective = True
                    break
        timer.stop('named expressions')
        timer.start('objective')
        if self._active_config.auto_updates.check_for_new_objective:
            pyomo_obj = get_objective(self._model)
            if pyomo_obj is not self._objective:
                need_to_set_objective = True
        else:
            pyomo_obj = self._objective
        if self._active_config.auto_updates.update_objective:
            if pyomo_obj is not None and pyomo_obj.expr is not self._objective_expr:
                need_to_set_objective = True
            elif pyomo_obj is not None and pyomo_obj.sense is not self._objective_sense:
                # we can definitely do something faster here than resetting the whole objective
                need_to_set_objective = True
        if need_to_set_objective:
            self.set_objective(pyomo_obj)
        timer.stop('objective')

        # this has to be done after the objective and constraints in case the
        # old objective/constraints use old variables
        timer.start('vars')
        self.remove_variables(old_vars)
        timer.stop('vars')


class PersistentSolverMixin:
    """
    The `solve` method in Gurobi and Highs is exactly the same, so this Mixin
    minimizes the duplicate code
    """

    def solve(self, model, **kwds) -> Results:
        start_timestamp = datetime.datetime.now(datetime.timezone.utc)
        tick = time.perf_counter()
        self._active_config = config = self.config(value=kwds, preserve_implicit=True)
        StaleFlagManager.mark_all_as_stale()

        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()
        if config.timer is None:
            config.timer = HierarchicalTimer()
        timer = config.timer

        if model is not self._model:
            timer.start('set_instance')
            self.set_instance(model)
            timer.stop('set_instance')
        else:
            timer.start('update')
            self.update(timer=timer)
            timer.stop('update')

        res = self._solve()
        self._last_results_object = res

        tock = time.perf_counter()
        res.timing_info.start_timestamp = start_timestamp
        res.timing_info.wall_time = tock - tick
        res.timing_info.timer = timer
        self._active_config = self.config

        return res



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import math

from pyomo.common.dependencies import attempt_import
from pyomo.core.base.var import ScalarVar
from pyomo.core.base.expression import ScalarExpression
import pyomo.core.expr.expr_common as common
import pyomo.core.expr as EXPR
from pyomo.core.expr.numvalue import (
    value,
    is_constant,
    is_fixed,
    native_numeric_types,
    native_types,
    nonpyomo_leaf_types,
)
from pyomo.core.kernel.objective import minimize, maximize
from pyomo.repn.util import valid_expr_ctypes_minlp

maingopy, maingopy_available = attempt_import("maingopy")

_plusMinusOne = {1, -1}

LEFT_TO_RIGHT = common.OperatorAssociativity.LEFT_TO_RIGHT
RIGHT_TO_LEFT = common.OperatorAssociativity.RIGHT_TO_LEFT


class ToMAiNGOVisitor(EXPR.ExpressionValueVisitor):
    def __init__(self, variables, idmap):
        super(ToMAiNGOVisitor, self).__init__()
        self.variables = variables
        self.idmap = idmap
        self._pyomo_func_to_maingo_func = {
            "log": maingopy.log,
            "log10": ToMAiNGOVisitor.maingo_log10,
            "sin": maingopy.sin,
            "cos": maingopy.cos,
            "tan": maingopy.tan,
            "cosh": maingopy.cosh,
            "sinh": maingopy.sinh,
            "tanh": maingopy.tanh,
            "asin": maingopy.asin,
            "acos": maingopy.acos,
            "atan": maingopy.atan,
            "exp": maingopy.exp,
            "sqrt": maingopy.sqrt,
            "asinh": ToMAiNGOVisitor.maingo_asinh,
            "acosh": ToMAiNGOVisitor.maingo_acosh,
            "atanh": ToMAiNGOVisitor.maingo_atanh,
        }

    @classmethod
    def maingo_log10(cls, x):
        return maingopy.log(x) / math.log(10)

    @classmethod
    def maingo_asinh(cls, x):
        return maingopy.log(x + maingopy.sqrt(maingopy.pow(x, 2) + 1))

    @classmethod
    def maingo_acosh(cls, x):
        return maingopy.log(x + maingopy.sqrt(maingopy.pow(x, 2) - 1))

    @classmethod
    def maingo_atanh(cls, x):
        return 0.5 * maingopy.log(x + 1) - 0.5 * maingopy.log(1 - x)

    def visit(self, node, values):
        """Visit nodes that have been expanded"""
        for i, val in enumerate(values):
            arg = node._args_[i]

            if arg is None:
                values[i] = "Undefined"
            elif arg.__class__ in native_numeric_types:
                pass
            elif arg.__class__ in nonpyomo_leaf_types:
                values[i] = val
            else:
                parens = False
                if arg.is_expression_type() and node.PRECEDENCE is not None:
                    if arg.PRECEDENCE is None:
                        pass
                    elif node.PRECEDENCE < arg.PRECEDENCE:
                        parens = True
                    elif node.PRECEDENCE == arg.PRECEDENCE:
                        if i == 0:
                            parens = node.ASSOCIATIVITY != LEFT_TO_RIGHT
                        elif i == len(node._args_) - 1:
                            parens = node.ASSOCIATIVITY != RIGHT_TO_LEFT
                        else:
                            parens = True
                if parens:
                    values[i] = val

        if node.__class__ in EXPR.NPV_expression_types:
            return value(node)

        if node.__class__ in {EXPR.ProductExpression, EXPR.MonomialTermExpression}:
            return values[0] * values[1]

        if node.__class__ in {EXPR.SumExpression}:
            return sum(values)

        if node.__class__ in {EXPR.PowExpression}:
            return maingopy.pow(values[0], values[1])

        if node.__class__ in {EXPR.DivisionExpression}:
            return values[0] / values[1]

        if node.__class__ in {EXPR.NegationExpression}:
            return -values[0]

        if node.__class__ in {EXPR.AbsExpression}:
            return maingopy.abs(values[0])

        if node.__class__ in {EXPR.UnaryFunctionExpression}:
            pyomo_func = node.getname()
            maingo_func = self._pyomo_func_to_maingo_func[pyomo_func]
            return maingo_func(values[0])

        if node.__class__ in {ScalarExpression}:
            return values[0]

        raise ValueError(f"Unknown function expression encountered: {node.getname()}")

    def visiting_potential_leaf(self, node):
        """
        Visiting a potential leaf.

        Return True if the node is not expanded.
        """
        if node.__class__ in native_types:
            return True, node

        if node.is_expression_type():
            if node.__class__ is EXPR.MonomialTermExpression:
                return True, self._monomial_to_maingo(node)
            if node.__class__ is EXPR.LinearExpression:
                return True, self._linear_to_maingo(node)
            return False, None

        if node.is_component_type():
            if node.ctype not in valid_expr_ctypes_minlp:
                # Make sure all components in active constraints
                # are basic ctypes we know how to deal with.
                raise RuntimeError(
                    "Unallowable component '%s' of type %s found in an active "
                    "constraint or objective.\nMAiNGO cannot export "
                    "expressions with this component type."
                    % (node.name, node.ctype.__name__)
                )

        if node.is_fixed():
            return True, node()
        else:
            assert node.is_variable_type()
            maingo_var_id = self.idmap[id(node)]
            maingo_var = self.variables[maingo_var_id]
            return True, maingo_var

    def _monomial_to_maingo(self, node):
        const, var = node.args
        if const.__class__ not in native_types:
            const = value(const)
        if var.is_fixed():
            return const * var.value
        if not const:
            return 0
        maingo_var = self._var_to_maingo(var)
        if const in _plusMinusOne:
            if const < 0:
                return -maingo_var
            else:
                return maingo_var
        return const * maingo_var

    def _var_to_maingo(self, var):
        maingo_var_id = self.idmap[id(var)]
        maingo_var = self.variables[maingo_var_id]
        return maingo_var

    def _linear_to_maingo(self, node):
        values = [
            (
                self._monomial_to_maingo(arg)
                if (arg.__class__ is EXPR.MonomialTermExpression)
                else (
                    value(arg)
                    if arg.__class__ in native_numeric_types
                    else (
                        self._var_to_maingo(arg)
                        if arg.is_variable_type()
                        else value(arg)
                    )
                )
            )
            for arg in node.args
        ]
        return sum(values)


class SolverModel(maingopy.MAiNGOmodel if maingopy_available else object):
    def __init__(self, var_list, objective, con_list, idmap, logger):
        maingopy.MAiNGOmodel.__init__(self)
        self._var_list = var_list
        self._con_list = con_list
        self._objective = objective
        self._idmap = idmap
        self._logger = logger
        self._no_objective = False

        if self._objective is None:
            self._logger.warning("No objective given, setting a dummy objective of 1.")
            self._no_objective = True

    def build_maingo_objective(self, obj, visitor):
        if self._no_objective:
            return visitor.variables[-1]
        maingo_obj = visitor.dfs_postorder_stack(obj.expr)
        if obj.sense == maximize:
            return -1 * maingo_obj
        return maingo_obj

    def build_maingo_constraints(self, cons, visitor):
        eqs = []
        ineqs = []
        for con in cons:
            if con.equality:
                eqs += [visitor.dfs_postorder_stack(con.body - con.lower)]
            elif con.has_ub() and con.has_lb():
                ineqs += [visitor.dfs_postorder_stack(con.body - con.upper)]
                ineqs += [visitor.dfs_postorder_stack(con.lower - con.body)]
            elif con.has_ub():
                ineqs += [visitor.dfs_postorder_stack(con.body - con.upper)]
            elif con.has_lb():
                ineqs += [visitor.dfs_postorder_stack(con.lower - con.body)]
            else:
                raise ValueError(
                    "Constraint does not have a lower "
                    "or an upper bound: {0} \n".format(con)
                )
        return eqs, ineqs

    def get_variables(self):
        vars = [
            maingopy.OptimizationVariable(
                maingopy.Bounds(var.lb, var.ub), var.type, var.name
            )
            for var in self._var_list
        ]
        if self._no_objective:
            vars += [maingopy.OptimizationVariable(maingopy.Bounds(1, 1), "dummy_obj")]
        return vars

    def get_initial_point(self):
        initial = [
            var.init if not var.init is None else (var.lb + var.ub) / 2.0
            for var in self._var_list
        ]
        if self._no_objective:
            initial += [1]
        return initial

    def evaluate(self, maingo_vars):
        visitor = ToMAiNGOVisitor(maingo_vars, self._idmap)
        result = maingopy.EvaluationContainer()
        result.objective = self.build_maingo_objective(self._objective, visitor)
        eqs, ineqs = self.build_maingo_constraints(self._con_list, visitor)
        result.eq = eqs
        result.ineq = ineqs
        return result



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.expr.visitor import ExpressionValueVisitor, nonpyomo_leaf_types
import pyomo.core.expr as EXPR


class _VarAndNamedExprCollector(ExpressionValueVisitor):
    def __init__(self):
        self.named_expressions = dict()
        self.variables = dict()
        self.fixed_vars = dict()
        self._external_functions = dict()

    def visit(self, node, values):
        pass

    def visiting_potential_leaf(self, node):
        if type(node) in nonpyomo_leaf_types:
            return True, None

        if node.is_variable_type():
            self.variables[id(node)] = node
            if node.is_fixed():
                self.fixed_vars[id(node)] = node
            return True, None

        if node.is_named_expression_type():
            self.named_expressions[id(node)] = node
            return False, None

        if type(node) is EXPR.ExternalFunctionExpression:
            self._external_functions[id(node)] = node
            return False, None

        if node.is_expression_type():
            return False, None

        return True, None


_visitor = _VarAndNamedExprCollector()


def collect_vars_and_named_exprs(expr):
    _visitor.__init__()
    _visitor.dfs_postorder_stack(expr)
    return (
        list(_visitor.named_expressions.values()),
        list(_visitor.variables.values()),
        list(_visitor.fixed_vars.values()),
        list(_visitor._external_functions.values()),
    )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo
from pyomo.repn import generate_standard_repn
import pyomo.common.unittest as unittest
from pyomo.common.collections import ComponentSet
from pyomo.contrib.incidence_analysis.incidence import (
    IncidenceMethod,
    get_incident_variables,
)


class TestAssumedBehavior(unittest.TestCase):
    """Tests for non-obvious behavior we rely on

    If this behavior changes, these tests will fail and hopefully we won't
    waste time debugging the "real" tests. This behavior includes:
    - The error message when we try to evaluate an expression with
      uninitialized variables

    """

    def test_uninitialized_value_error_message(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2])
        m.x[1].set_value(5)
        msg = "No value for uninitialized VarData"
        with self.assertRaisesRegex(ValueError, msg):
            pyo.value(1 + m.x[1] * m.x[2])


class _TestIncidence:
    """Base class with tests for get_incident_variables that should be
    independent of the method used

    """

    def _get_incident_variables(self, expr):
        raise NotImplementedError("_TestIncidence should not be used directly")

    def test_basic_incidence(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3])
        expr = m.x[1] + m.x[1] * m.x[2] + m.x[1] * pyo.exp(m.x[3])
        variables = self._get_incident_variables(expr)
        self.assertEqual(ComponentSet(variables), ComponentSet(m.x[:]))

    def test_incidence_with_fixed_variable(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3], initialize=1.0)
        expr = m.x[1] + m.x[1] * m.x[2] + m.x[1] * pyo.exp(m.x[3])
        m.x[2].fix()
        variables = self._get_incident_variables(expr)
        var_set = ComponentSet(variables)
        self.assertEqual(var_set, ComponentSet([m.x[1], m.x[3]]))

    def test_incidence_with_named_expression(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3])
        m.subexpr = pyo.Expression(pyo.Integers)
        m.subexpr[1] = m.x[1] * pyo.exp(m.x[3])
        expr = m.x[1] + m.x[1] * m.x[2] + m.subexpr[1]
        variables = self._get_incident_variables(expr)
        self.assertEqual(ComponentSet(variables), ComponentSet(m.x[:]))


class _TestIncidenceLinearOnly:
    """Tests for methods that support linear_only"""

    def _get_incident_variables(self, expr):
        raise NotImplementedError(
            "_TestIncidenceLinearOnly should not be used directly"
        )

    def test_linear_only(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3])

        expr = 2 * m.x[1] + 4 * m.x[2] * m.x[1] - m.x[1] * pyo.exp(m.x[3])
        variables = self._get_incident_variables(expr, linear_only=True)
        self.assertEqual(len(variables), 0)

        expr = 2 * m.x[1] + 2 * m.x[2] * m.x[3] + 3 * m.x[2]
        variables = self._get_incident_variables(expr, linear_only=True)
        self.assertEqual(ComponentSet(variables), ComponentSet([m.x[1]]))

        m.x[3].fix(2.5)
        expr = 2 * m.x[1] + 2 * m.x[2] * m.x[3] + 3 * m.x[2]
        variables = self._get_incident_variables(expr, linear_only=True)
        self.assertEqual(ComponentSet(variables), ComponentSet([m.x[1], m.x[2]]))


class _TestIncidenceLinearCancellation:
    """Tests for methods that perform linear cancellation"""

    def _get_incident_variables(self, expr):
        raise NotImplementedError(
            "_TestIncidenceLinearCancellation should not be used directly"
        )

    def test_zero_coef(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3])

        # generate_standard_repn filters subexpressions with zero coefficients
        expr = 0 * m.x[1] + 0 * m.x[1] * m.x[2] + 0 * pyo.exp(m.x[3])
        variables = self._get_incident_variables(expr)
        self.assertEqual(len(variables), 0)

    def test_variable_minus_itself(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3])
        # standard repn will recognize the zero coefficient and filter x[1]
        expr = m.x[1] + m.x[2] * m.x[3] - m.x[1]
        variables = self._get_incident_variables(expr)
        var_set = ComponentSet(variables)
        self.assertEqual(var_set, ComponentSet([m.x[2], m.x[3]]))

    def test_fixed_zero_linear_coefficient(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3])
        m.p = pyo.Param([1, 2], mutable=True, initialize=1.0)
        m.p[1].set_value(0)
        expr = 2 * m.x[1] + m.p[1] * m.p[2] * m.x[2] + m.p[2] * m.x[3] ** 2
        variables = self._get_incident_variables(expr)
        self.assertEqual(ComponentSet(variables), ComponentSet([m.x[1], m.x[3]]))

        m.x[3].fix(0.0)
        expr = 2 * m.x[1] + 3 * m.x[3] * m.p[2] * m.x[2] + m.x[1] ** 2
        variables = self._get_incident_variables(expr)
        self.assertEqual(ComponentSet(variables), ComponentSet([m.x[1]]))

        m.x[3].fix(1.0)
        variables = self._get_incident_variables(expr)
        self.assertEqual(ComponentSet(variables), ComponentSet([m.x[1], m.x[2]]))

    # NOTE: This test assumes that all methods that support linear cancellation
    # accept a linear_only argument. If this changes, this test will need to be
    # moved.
    def test_fixed_zero_coefficient_linear_only(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3])
        expr = m.x[1] * m.x[2] + 2 * m.x[3]
        m.x[2].fix(0)
        variables = get_incident_variables(
            expr, method=IncidenceMethod.standard_repn, linear_only=True
        )
        self.assertEqual(len(variables), 1)
        self.assertIs(variables[0], m.x[3])


class TestIncidenceStandardRepn(
    unittest.TestCase,
    _TestIncidence,
    _TestIncidenceLinearOnly,
    _TestIncidenceLinearCancellation,
):
    def _get_incident_variables(self, expr, **kwds):
        method = IncidenceMethod.standard_repn
        return get_incident_variables(expr, method=method, **kwds)

    def test_assumed_standard_repn_behavior(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2])
        m.p = pyo.Param(initialize=0.0)

        # We rely on variables with constant coefficients of zero not appearing
        # in the standard repn (as opposed to appearing with explicit
        # coefficients of zero).
        expr = m.x[1] + 0 * m.x[2]
        repn = generate_standard_repn(expr)
        self.assertEqual(len(repn.linear_vars), 1)
        self.assertIs(repn.linear_vars[0], m.x[1])

        expr = m.p * m.x[1] + m.x[2]
        repn = generate_standard_repn(expr)
        self.assertEqual(len(repn.linear_vars), 1)
        self.assertIs(repn.linear_vars[0], m.x[2])

    def test_fixed_none_linear_coefficient(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3])
        m.p = pyo.Param([1, 2], mutable=True, initialize=1.0)
        m.x[3].fix(None)
        expr = 2 * m.x[1] + 3 * m.x[3] * m.p[2] * m.x[2] + m.x[1] ** 2
        variables = self._get_incident_variables(expr)
        self.assertEqual(ComponentSet(variables), ComponentSet([m.x[1], m.x[2]]))

    def test_incidence_with_mutable_parameter(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3])
        m.p = pyo.Param(mutable=True, initialize=None)
        expr = m.x[1] + m.p * m.x[1] * m.x[2] + m.x[1] * pyo.exp(m.x[3])
        variables = self._get_incident_variables(expr)
        self.assertEqual(ComponentSet(variables), ComponentSet(m.x[:]))


class TestIncidenceIdentifyVariables(unittest.TestCase, _TestIncidence):
    def _get_incident_variables(self, expr, **kwds):
        method = IncidenceMethod.identify_variables
        return get_incident_variables(expr, method=method, **kwds)

    def test_zero_coef(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3])

        # identify_variables does not eliminate expressions times zero
        expr = 0 * m.x[1] + 0 * m.x[1] * m.x[2] + 0 * pyo.exp(m.x[3])
        variables = self._get_incident_variables(expr)
        self.assertEqual(ComponentSet(variables), ComponentSet(m.x[:]))

    def test_variable_minus_itself(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3])
        # identify_variables will not filter x[1]
        expr = m.x[1] + m.x[2] * m.x[3] - m.x[1]
        variables = self._get_incident_variables(expr)
        var_set = ComponentSet(variables)
        self.assertEqual(var_set, ComponentSet(m.x[:]))

    def test_incidence_with_mutable_parameter(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3])
        m.p = pyo.Param(mutable=True, initialize=None)
        expr = m.x[1] + m.p * m.x[1] * m.x[2] + m.x[1] * pyo.exp(m.x[3])
        variables = self._get_incident_variables(expr)
        self.assertEqual(ComponentSet(variables), ComponentSet(m.x[:]))


class TestIncidenceAmplRepn(
    unittest.TestCase,
    _TestIncidence,
    _TestIncidenceLinearOnly,
    _TestIncidenceLinearCancellation,
):
    def _get_incident_variables(self, expr, **kwds):
        method = IncidenceMethod.ampl_repn
        return get_incident_variables(expr, method=method, **kwds)


class TestIncidenceStandardRepnComputeValues(
    unittest.TestCase,
    _TestIncidence,
    _TestIncidenceLinearOnly,
    _TestIncidenceLinearCancellation,
):
    def _get_incident_variables(self, expr, **kwds):
        method = IncidenceMethod.standard_repn_compute_values
        return get_incident_variables(expr, method=method, **kwds)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

logger = logging.getLogger(__name__)

from contextlib import contextmanager

from pyomo.common.dependencies import numpy as numpy, numpy_available

if numpy_available:
    import numpy.random
    from numpy.linalg import norm

import pyomo.environ as pyo
from pyomo.common.modeling import unique_component_name
from pyomo.common.collections import ComponentSet
import pyomo.util.vars_from_expressions as vfe


@contextmanager
def logcontext(level):
    """
    This context manager is used to dynamically set the specified logging level
    and then execute a block of code using that logging level.  When the context is
    deleted, the logging level is reset to the original value.

    Examples
    --------
    >>> with logcontext(logging.INFO):
    ...    logging.debug("This will not be printed")
    ...    logging.info("This will be printed")

    """
    logger = logging.getLogger()
    current_level = logger.getEffectiveLevel()
    logger.setLevel(level)
    try:
        yield
    finally:
        logger.setLevel(current_level)


def get_active_objective(model):
    """
    Finds and returns the active objective function for a model. Currently
    assume that there is exactly one active objective.
    """

    active_objs = list(model.component_data_objects(pyo.Objective, active=True))
    assert (
        len(active_objs) == 1
    ), "Model has {} active objective functions, exactly one is required.".format(
        len(active_objs)
    )

    return active_objs[0]


def _add_aos_block(model, name="_aos_block"):
    """Adds an alternative optimal solution block with a unique name."""
    aos_block = pyo.Block()
    model.add_component(unique_component_name(model, name), aos_block)
    return aos_block


def _add_objective_constraint(
    aos_block, objective, objective_value, rel_opt_gap, abs_opt_gap
):
    """
    Adds a relative and/or absolute objective function constraint to the
    specified block.
    """

    assert (
        rel_opt_gap is None or rel_opt_gap >= 0.0
    ), "rel_opt_gap must be None or >= 0.0"
    assert (
        abs_opt_gap is None or abs_opt_gap >= 0.0
    ), "abs_opt_gap must be None or >= 0.0"

    objective_constraints = []

    objective_is_min = objective.is_minimizing()
    objective_expr = objective.expr

    objective_sense = -1
    if objective_is_min:
        objective_sense = 1

    if rel_opt_gap is not None:
        objective_cutoff = objective_value + objective_sense * rel_opt_gap * abs(
            objective_value
        )

        if objective_is_min:
            aos_block.optimality_tol_rel = pyo.Constraint(
                expr=objective_expr <= objective_cutoff
            )
        else:
            aos_block.optimality_tol_rel = pyo.Constraint(
                expr=objective_expr >= objective_cutoff
            )
        objective_constraints.append(aos_block.optimality_tol_rel)

    if abs_opt_gap is not None:
        objective_cutoff = objective_value + objective_sense * abs_opt_gap

        if objective_is_min:
            aos_block.optimality_tol_abs = pyo.Constraint(
                expr=objective_expr <= objective_cutoff
            )
        else:
            aos_block.optimality_tol_abs = pyo.Constraint(
                expr=objective_expr >= objective_cutoff
            )
        objective_constraints.append(aos_block.optimality_tol_abs)

    return objective_constraints


if numpy_available:
    rng = numpy.random.default_rng(9283749387)
else:
    rng = None


def _set_numpy_rng(seed):
    global rng
    rng = numpy.random.default_rng(seed)


def _get_random_direction(num_dimensions, iterations=1000, min_norm=1e-4):
    """
    Get a unit vector of dimension num_dimensions by sampling from and
    normalizing a standard multivariate Gaussian distribution.
    """
    for idx in range(iterations):
        samples = rng.normal(size=num_dimensions)
        samples_norm = norm(samples)
        if samples_norm > min_norm:
            return samples / samples_norm
    raise Exception(  # pragma: no cover
        (
            "Generated {} sequential Gaussian draws with a norm of "
            "less than {}.".format(iterations, min_norm)
        )
    )


def _filter_model_variables(
    variable_set,
    var_generator,
    include_continuous=True,
    include_binary=True,
    include_integer=True,
    include_fixed=False,
):
    """
    Filters variables from a variable generator and adds them to a set.
    """
    for var in var_generator:
        if var in variable_set or (var.is_fixed() and not include_fixed):
            continue
        if (
            (var.is_continuous() and include_continuous)
            or (var.is_binary() and include_binary)
            or (var.is_integer() and include_integer)
        ):
            variable_set.add(var)


def get_model_variables(
    model,
    components=None,
    include_continuous=True,
    include_binary=True,
    include_integer=True,
    include_fixed=False,
):
    """Gathers and returns all variables or a subset of variables from a
    Pyomo model.

    Parameters
    ----------
    model : ConcreteModel
        A concrete Pyomo model.
    components: None or a collection of Pyomo components
        The components from which variables should be collected. None
        indicates that all variables will be included. Alternatively, a
        collection of Pyomo Blocks, Constraints, or Variables (indexed or
        non-indexed) from which variables will be gathered can be provided.
        If a Block is provided, all variables associated with constraints
        in that that block and its sub-blocks will be returned. To exclude
        sub-blocks, a tuple element with the format (Block, False) can be
        used.
    include_continuous : boolean
        Boolean indicating that continuous variables should be included.
    include_binary : boolean
        Boolean indicating that binary variables should be included.
    include_integer : boolean
        Boolean indicating that integer variables should be included.
    include_fixed : boolean
        Boolean indicating that fixed variables should be included.

    Returns
    -------
    variable_set
        A Pyomo ComponentSet containing _GeneralVarData variables.

    """

    component_list = (pyo.Objective, pyo.Constraint)
    variable_set = ComponentSet()
    if components == None:
        var_generator = vfe.get_vars_from_components(
            model, component_list, include_fixed=include_fixed
        )
        _filter_model_variables(
            variable_set,
            var_generator,
            include_continuous,
            include_binary,
            include_integer,
            include_fixed,
        )
    else:
        for comp in components:
            if hasattr(comp, "ctype") and comp.ctype == pyo.Block:
                blocks = comp.values() if comp.is_indexed() else (comp,)
                for item in blocks:
                    variables = vfe.get_vars_from_components(
                        item, component_list, include_fixed=include_fixed
                    )
                    _filter_model_variables(
                        variable_set,
                        variables,
                        include_continuous,
                        include_binary,
                        include_integer,
                        include_fixed,
                    )
            elif (
                isinstance(comp, tuple)
                and hasattr(comp[0], "ctype")
                and comp[0].ctype == pyo.Block
            ):
                block = comp[0]
                descend_into = pyo.Block if comp[1] else False
                blocks = block.values() if block.is_indexed() else (block,)
                for item in blocks:
                    variables = vfe.get_vars_from_components(
                        item,
                        component_list,
                        include_fixed=include_fixed,
                        descend_into=descend_into,
                    )
                    _filter_model_variables(
                        variable_set,
                        variables,
                        include_continuous,
                        include_binary,
                        include_integer,
                        include_fixed,
                    )
            elif hasattr(comp, "ctype") and comp.ctype in component_list:
                constraints = comp.values() if comp.is_indexed() else (comp,)
                for item in constraints:
                    variables = pyo.expr.identify_variables(
                        item.expr, include_fixed=include_fixed
                    )
                    _filter_model_variables(
                        variable_set,
                        variables,
                        include_continuous,
                        include_binary,
                        include_integer,
                        include_fixed,
                    )
            elif hasattr(comp, "ctype") and comp.ctype == pyo.Var:
                variables = comp.values() if comp.is_indexed() else (comp,)
                _filter_model_variables(
                    variable_set,
                    variables,
                    include_continuous,
                    include_binary,
                    include_integer,
                    include_fixed,
                )
            else:  # pragma: no cover
                logger.info(
                    ("No variables added for unrecognized component {}.").format(comp)
                )

    return variable_set
